#
# Gramps - a GTK+/GNOME based genealogy program
#
# Copyright (C) 2016 Tom Samstag
# Copyright (C) 2025 Doug Blank <doug.blank@gmail.com>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#

"""
Unittest that tests person-specific filter rules
"""
import unittest
import os
from time import perf_counter
import inspect
from contextlib import contextmanager

from ....filters import reload_custom_filters

reload_custom_filters()
from ....db.utils import import_as_dict
from ....filters import GenericFilter, CustomFilters
from ....const import DATA_DIR
from ....user import User
from ....utils.unittest import localize_date

from ..person import (
    Disconnected,
    Everyone,
    FamilyWithIncompleteEvent,
    HasAddress,
    HasAlternateName,
    HasAssociation,
    HasBirth,
    HasDeath,
    HasEvent,
    HasCommonAncestorWith,
    HasCommonAncestorWithFilterMatch,
    HasFamilyAttribute,
    HasFamilyEvent,
    HasIdOf,
    HasLDS,
    HasNameOf,
    HasNameOriginType,
    HasNameType,
    HasNickname,
    HasRelationship,
    HasSoundexName,
    HasSourceOf,
    HasTextMatchingRegexpOf,
    HasUnknownGender,
    HaveAltFamilies,
    HaveChildren,
    HavePhotos,
    IncompleteNames,
    IsAncestorOfFilterMatch,
    IsBookmarked,
    IsChildOfFilterMatch,
    IsDescendantFamilyOf,
    IsDescendantFamilyOfFilterMatch,
    IsDescendantOfFilterMatch,
    IsDefaultPerson,
    IsDescendantOf,
    IsDuplicatedAncestorOf,
    IsFemale,
    IsLessThanNthGenerationAncestorOf,
    IsLessThanNthGenerationAncestorOfDefaultPerson,
    IsLessThanNthGenerationAncestorOfBookmarked,
    IsMale,
    IsMoreThanNthGenerationAncestorOf,
    IsMoreThanNthGenerationDescendantOf,
    IsParentOfFilterMatch,
    IsRelatedWith,
    IsSiblingOfFilterMatch,
    IsSpouseOfFilterMatch,
    IsWitness,
    MissingParent,
    MultipleMarriages,
    NeverMarried,
    NoBirthdate,
    NoDeathdate,
    PeoplePrivate,
    PeoplePublic,
    PersonWithIncompleteEvent,
    ProbablyAlive,
    RegExpName,
    RelationshipPathBetweenBookmarks,
)

TEST_DIR = os.path.abspath(os.path.join(DATA_DIR, "tests"))
EXAMPLE = os.path.join(TEST_DIR, "example.gramps")


@contextmanager
def count_method_calls(cls, method_name):
    """
    Context manager to monkey-patch a method and count its calls.

    Args:
        cls: The class containing the method.
        method_name: The name of the method to patch.

    Yields:
        A callable that returns the call count.
    """
    original_method = getattr(cls, method_name)
    call_count = 0

    def patched_method(self, *args, **kwargs):
        nonlocal call_count
        call_count += 1
        return original_method(self, *args, **kwargs)

    setattr(cls, method_name, patched_method)

    def get_call_count():
        return call_count

    try:
        yield get_call_count
    finally:
        setattr(cls, method_name, original_method)


class BaseTest(unittest.TestCase):
    """
    Person rule tests.
    """

    @classmethod
    def setUpClass(cls):
        """
        Import example database.
        """
        cls.db = import_as_dict(EXAMPLE, User())

    def filter_with_rule(
        self,
        rule,
        l_op="and",
        invert=False,
        baserule=None,
        base_l_op="and",
        base_invert=False,
        base_name="Base",
    ):
        """
        Apply a filter with the given rule.  'baserule' can be used to stack
        filters when using filters that are of 'offiltermatch' type.
        """
        if baserule:
            filter_ = GenericFilter()
            if isinstance(baserule, list):
                filter_.set_rules(baserule)
            else:
                filter_.add_rule(baserule)
            filter_.set_logical_op(base_l_op)
            filter_.set_invert(base_invert)
            filter_.set_name(base_name)
            filters = CustomFilters.get_filters_dict("Person")
            filters[base_name] = filter_
        filter_ = GenericFilter()
        if isinstance(rule, list):
            filter_.set_rules(rule)
        else:
            filter_.add_rule(rule)
        filter_.set_logical_op(l_op)
        filter_.set_invert(invert)
        stime = perf_counter()
        results = filter_.apply(self.db)
        # if __debug__:
        # frame = inspect.currentframe()
        # rulename = frame.f_back.f_code.co_name
        # print("%s: %.2f\n" % (rulename, perf_counter() - stime))
        return set(results)

    def test_Complex_1(self):
        """Test with two ancestor trees in base filter, and a complex
        'or' of a descendent tree, sibling of,"""
        rule1 = IsLessThanNthGenerationAncestorOf(["I0005", 10])
        rule2 = IsLessThanNthGenerationAncestorOf(["I0006", 10])
        ruleA = IsDescendantOfFilterMatch(["Base"])
        ruleB = IsSiblingOfFilterMatch(["Base"])
        ruleC = IsSpouseOfFilterMatch(["Base"])
        res = self.filter_with_rule(
            [ruleA, ruleB, ruleC], l_op="or", baserule=[rule1, rule2], base_l_op="or"
        )
        self.assertEqual(len(res), 1277)

    def test_IsDescendantOfFilterMatch(self):
        """Test the rule with two persons in base filter"""
        rule1 = HasIdOf(["I0610"])
        rule2 = HasIdOf(["I1804"])
        rule = IsDescendantOfFilterMatch(["Base"])
        res = self.filter_with_rule(rule, baserule=[rule1, rule2], base_l_op="or")
        self.assertEqual(len(res), 120)

    def test_IsDescendantFamilyOfFilterMatch(self):
        """Test the rule with two persons in base filter"""
        rule1 = HasIdOf(["I0610"])
        rule2 = HasIdOf(["I1804"])
        rule = IsDescendantFamilyOfFilterMatch(["Base"])
        res = self.filter_with_rule(rule, baserule=[rule1, rule2], base_l_op="or")
        self.assertEqual(len(res), 171)

    def test_IsDescendantFamilyOf(self):
        """Test the rule"""
        rule = IsDescendantFamilyOf(["I0610", 1])
        res = self.filter_with_rule(rule)
        self.assertEqual(len(res), 118)

    def test_IsDescendantOf(self):
        """Test the rule"""
        rule = IsDescendantOf(["I0610", 0])
        res = self.filter_with_rule(rule)
        self.assertEqual(len(res), 85)

    def test_IsMoreThanNthGenerationDescendantOf(self):
        """Test the rule"""
        rule = IsMoreThanNthGenerationDescendantOf(["I0610", 3])
        res = self.filter_with_rule(rule)
        self.assertEqual(len(res), 83)

    def test_IsLessThanNthGenerationAncestorOfDefaultPerson(self):
        """Test the rule"""
        rule = IsLessThanNthGenerationAncestorOfDefaultPerson([5])
        res = self.filter_with_rule(rule)
        self.assertEqual(len(res), 7)

    def test_IsLessThanNthGenerationAncestorOfBookmarked(self):
        """Test the rule"""
        rule = IsLessThanNthGenerationAncestorOfBookmarked([5])
        res = self.filter_with_rule(rule)
        self.assertEqual(len(res), 8)

    def test_IsParentOfFilterMatch(self):
        """Test the rule with two persons in base filter"""
        rule1 = HasIdOf(["I0006"])
        rule2 = HasIdOf(["I0010"])
        rule = IsParentOfFilterMatch(["Base"])
        res = self.filter_with_rule(rule, baserule=[rule1, rule2], base_l_op="or")
        self.assertEqual(len(res), 4)

    def test_IsSiblingOfFilterMatch(self):
        """Test the rule with two persons in base filter"""
        rule1 = HasIdOf(["I0006"])
        rule2 = HasIdOf(["I0010"])
        rule = IsSiblingOfFilterMatch(["Base"])
        res = self.filter_with_rule(rule, baserule=[rule1, rule2], base_l_op="or")
        self.assertEqual(len(res), 7)

    def test_IsSpouseOfFilterMatch(self):
        """Test the rule with two persons in base filter"""
        rule1 = HasIdOf(["I0006"])
        rule2 = HasIdOf(["I0010"])
        rule = IsSpouseOfFilterMatch(["Base"])
        res = self.filter_with_rule(rule, baserule=[rule1, rule2], base_l_op="or")
        self.assertEqual(len(res), 2)

    def test_isancestoroffiltermatch(self):
        """Test the rule with two persons in base filter"""
        rule1 = HasIdOf(["I0006"])
        rule2 = HasIdOf(["I0005"])
        rule = IsAncestorOfFilterMatch(["Base"])
        res = self.filter_with_rule(rule, baserule=[rule1, rule2], base_l_op="or")
        self.assertEqual(len(res), 431)

    def test_HasCommonAncestorWithFilterMatch(self):
        """Test the rule with two persons in base filter"""
        rule1 = HasIdOf(["I0006"])
        rule2 = HasIdOf(["I0005"])
        rule = HasCommonAncestorWithFilterMatch(["Base"])
        res = self.filter_with_rule(rule, baserule=[rule1, rule2], base_l_op="or")
        self.assertEqual(len(res), 1415)

    def test_IsChildOfFilterMatch(self):
        """Test the rule with two persons in base filter"""
        rule1 = HasIdOf(["I0006"])
        rule2 = HasIdOf(["I0010"])
        rule = IsChildOfFilterMatch(["Base"])
        res = self.filter_with_rule(rule, baserule=[rule1, rule2], base_l_op="or")
        self.assertEqual(len(res), 11)

    def test_HasAddress(self):
        """
        Test HasAddress rule.
        """
        rule = HasAddress([0, "greater than"])
        res = self.filter_with_rule(rule)
        self.assertEqual(len(res), 1)

    def test_HasAssociation(self):
        """
        Test rule.
        """
        rule = HasAssociation([0, "greater than"])
        res = self.filter_with_rule(rule)
        self.assertEqual(len(res), 2)

    def test_HasBirth(self):
        """
        Test rule.
        """
        date_str = localize_date("between 1600 and 1700")
        rule = HasBirth([date_str, "akron", ""])
        res = self.filter_with_rule(rule)
        self.assertEqual(len(res), 2)

    def test_HasDeath(self):
        """
        Test HasDeath rule.
        """
        date_str = localize_date("between 1600 and 1700")
        rule = HasDeath([date_str, "ashtabula", ""])
        res = self.filter_with_rule(rule)
        self.assertEqual(len(res), 2)

    def test_HasEvent(self):
        """
        Test rule.
        """
        date_str = localize_date("between 1600 and 1700")
        rule = HasEvent(["Birth", date_str, "akron", "", "", 1])
        res = self.filter_with_rule(rule)
        self.assertEqual(len(res), 2)

    def test_HasFamilyAttribute(self):
        """
        Test rule.
        """
        rule = HasFamilyAttribute([5, "8"])
        res = self.filter_with_rule(rule)
        self.assertEqual(len(res), 2)

    def test_HasFamilyEvent(self):
        """
        Test rule.
        """
        date_str = localize_date("after 1900")
        rule = HasFamilyEvent(["Marriage", date_str, "craw", ""])
        res = self.filter_with_rule(rule)
        self.assertEqual(len(res), 4)

    def test_HavePhotos(self):
        """
        Test rule.
        """
        rule = HavePhotos([0, "greater than"])
        res = self.filter_with_rule(rule)
        self.assertEqual(len(res), 5)

    def test_HasLDS(self):
        """
        Test rule.
        """
        rule = HasLDS([0, "greater than"])
        res = self.filter_with_rule(rule)
        self.assertEqual(len(res), 1)

    def test_HasNameOriginType(self):
        """
        Test rule.
        """
        rule = HasNameOriginType(["Patrilineal"])
        res = self.filter_with_rule(rule)
        self.assertEqual(len(res), 9)

    def test_HasNameType(self):
        """
        Test rule.
        """
        rule = HasNameType(["Married Name"])
        res = self.filter_with_rule(rule)
        self.assertEqual(len(res), 1)

    def test_HasRelationship(self):
        """
        Test rule.
        """
        rule = HasRelationship([0, "Married", 0])
        res = self.filter_with_rule(rule)
        self.assertEqual(len(res), 1377)

    def test_HasTextMatchingRegexpOf(self):
        """
        Test rule.
        """
        rule = HasTextMatchingRegexpOf(
            [".*(Dahl|Akron|Smith|Attic|" "of Lessard).*", False], use_regex=True
        )
        res = self.filter_with_rule(rule)
        self.assertEqual(len(res), 28)

    def test_IsMoreThanNthGenerationAncestorOf(self):
        """
        Test rule.
        """
        rule = IsMoreThanNthGenerationAncestorOf(["I0005", 3])
        res = self.filter_with_rule(rule)
        self.assertEqual(len(res), 322)

    def test_IsWitness(self):
        """
        Test rule.
        """
        rule = IsWitness(["Marriage"])
        res = self.filter_with_rule(rule)
        self.assertEqual(len(res), 1)

    def test_ProbablyAlive(self):
        """
        Test rule.
        """
        rule = ProbablyAlive(["1900"])
        res = self.filter_with_rule(rule)
        self.assertEqual(len(res), 897)

    def test_RegExpName(self):
        """
        Test rule.
        """
        rule = RegExpName([".*(Garc|Amy).*"], use_regex=True)
        res = self.filter_with_rule(rule)
        self.assertEqual(len(res), 3)

    def test_disconnected(self):
        """
        Test Disconnected rule.
        """
        rule = Disconnected([])
        self.assertEqual(
            self.filter_with_rule(rule),
            set(
                [
                    "0PBKQCXHLAEIB46ZIA",
                    "QEVJQC04YO01UAWJ2N",
                    "UT0KQCMN7PC9XURRXJ",
                    "MZAKQCKAQLIQYWP5IW",
                    "Y7BKQC9CUXWQLGLPQM",
                    "OBBKQC8NJM5UYBO849",
                    "NPBKQCKEF0G7T4H312",
                    "423KQCGLT8UISDUM1Q",
                    "8S0KQCNORIWDL0X8SB",
                    "AP5KQC0LBXPM727OWB",
                    "AREKQC0VPBHNZ5R3IO",
                    "KU0KQCJ0RUTJTIUKSA",
                    "VC4KQC7L7KKH9RLHXN",
                    "0P3KQCRSIVL1A4VJ19",
                    "PK6KQCGEL4PTE720BL",
                    "YIKKQCSD2Z85UHJ8LX",
                    "KY8KQCMIH2HUUGLA3R",
                    "RD7KQCQ24B1N3OEC5X",
                    "NV0KQC7SIEH3SVDPP1",
                    "KIKKQCU2CJ543TLM5J",
                    "AT0KQC4P3MMUCHI3BK",
                    "J6BKQC1PMNBAYSLM9U",
                    "IXXJQCLKOUAJ5RSQY4",
                    "U4ZJQC5VR0QBIE8DU",
                    "F7BKQC4NXO9R7XOG2W",
                    "7U0KQC6PGZBNQATNOT",
                    "78AKQCI05U36T3E82O",
                    "H1GKQCWOUJHFSHXABA",
                    "ZWGKQCRFZAPC5PYJZ1",
                    "EZ0KQCF3LSM9PRSG0K",
                    "FHKKQC963NGSY18ZDZ",
                    "FJ9KQCRJ3RGHNBWW4S",
                    "S2EKQC9F4UR4R71IC3",
                    "1XBKQCX019BKJ0M9IH",
                    "Z62KQC706L0B0WTN3Q",
                    "O7EKQCEVZ7FBEWMNWE",
                    "XY8KQCULFPN4SR915Q",
                    "WQDKQCEULSD5G9XNFI",
                    "2Z0KQCSWKVFG7RPFD8",
                    "26BKQC0SJIJOH02H2A",
                    "262KQCH2RQKN0CBRLF",
                    "P5ZJQCMKO7EYV4HFCL",
                    "KXBKQC52JO3AP4GMLF",
                    "9IFKQC60JTDBV57N6S",
                    "TQ0KQCZ8LA7X9DIEAN",
                    "BAXJQCORQA5Q46FCDG",
                    "VR0KQC7LVANO83AL35",
                    "75CKQC4T617U2E5T5Y",
                    "LCTKQCZU3F94CEFSOM",
                    "WJYJQCPNJJI5JN07SD",
                    "3N6KQC6BE5EIXTRMDL",
                    "CM5KQCD57I15GKLAMB",
                    "cccbffffd3e69819cd8",
                    "BJKKQCVDA66528PDAU",
                    "QS0KQCLMIZFI8ZDLM3",
                    "UW0KQCRHBIYMA8LPZD",
                    "GJ7KQC7APJSAMHEK5Q",
                    "711KQCDXOQWB3KDWEP",
                    "PY0KQC77AJ3457A6C2",
                    "WZ0KQCYVMEJHDR4MV2",
                    "28EKQCQGM6NLLWFRG7",
                    "E33KQCRREJALRA715H",
                    "8HKKQCTEJAOBVH410L",
                    "IO6KQC70PMBQUDNB3L",
                    "1YBKQCWRBNB433NEMH",
                    "M01KQCF7KUWCDY67JD",
                    "CR0KQCOMV2QPPC90IF",
                    "85ZJQCMG38N7Q2WKIK",
                    "I9GKQCERACL8UZF2PY",
                    "BY0KQCOZUK47R2JZDE",
                    "7W0KQCYDMD4LTSY5JL",
                    "A0YJQC3HONEKD1JCPK",
                    "d5839c13b0541b7b8e6",
                ]
            ),
        )

    def test_everyone(self):
        """
        Test Everyone rule.
        """
        rule = Everyone([])
        self.assertEqual(
            len(self.filter_with_rule(rule)), self.db.get_number_of_people()
        )

    def test_familywithincompleteevent(self):
        """
        Test FamilyWithIncompleteEvent rule.
        """
        rule = FamilyWithIncompleteEvent([])
        # too many to list out to test explicitly
        self.assertEqual(len(self.filter_with_rule(rule)), 775)

    def test_hasalternatename(self):
        """
        Test HasAlternateName rule.
        """
        rule = HasAlternateName([])
        self.assertEqual(
            self.filter_with_rule(rule),
            set(
                [
                    "46WJQCIOLQ0KOX2XCC",
                    "GNUJQCL9MD64AM56OH",
                ]
            ),
        )

    def test_commonancestor_empty(self):
        """
        Test empty HasCommonAncestorWith rule.
        """
        rule = HasCommonAncestorWith([""])
        self.assertEqual(self.filter_with_rule(rule), set([]))

    def test_commonancestor_nonmatching(self):
        """
        Test non-matching HasCommonAncestorWith rule.
        """
        rule = HasCommonAncestorWith(["I0000"])
        self.assertEqual(self.filter_with_rule(rule), set(["d5839c1237765987724"]))

    def test_commonancestor_irregular(self):
        """
        Test irregular HasCommonAncestorWith rule.
        """
        rule = HasCommonAncestorWith(["ABCDEFG"])
        self.assertEqual(self.filter_with_rule(rule), set([]))

    def test_commonancestor_matching(self):
        """
        Test matching HasCommonAncestorWith rule.
        """
        rule = HasCommonAncestorWith(["I0044"])
        self.assertEqual(
            self.filter_with_rule(rule),
            set(
                [
                    "GNUJQCL9MD64AM56OH",
                    "SOFKQCBYAO18OWC0CS",
                    "EMEKQC02EOUF8H0SHM",
                    "3EXJQCVWOSQFGLYB6H",
                    "EMTJQCQU6TL4WAVSE4",
                    "QUEKQCZL61S8BJJ388",
                    "MKEKQCSBQGAVHAPCQT",
                    "MUFKQCMXUJ07MCDUNI",
                    "DBXJQCJCEZMO17WZ89",
                    "ORFKQC4KLWEGTGR19L",
                    "MG5KQC6ZKSVO4A63G2",
                    "N26KQCF3ASHMZ0HEW6",
                    "GNWJQC9NLVF2MZLHU9",
                    "ZFXJQCHAD8SLZZ7KRP",
                    "44WJQCLCQIPZUB0UH",
                    "B8TJQC53HJXOGXK8F7",
                    "D3WJQCCGV58IP8PNHZ",
                    "3LEKQCRF3FD2E1H73I",
                    "F06KQCZY1I4H4IFZM",
                    "VMTJQC49IGKLG2EQ5",
                    "9BXKQC1PVLPYFMD6IX",
                    "H1DKQC4YGZ5A61FGS",
                    "1GWJQCGOOZ8FJW3YK9",
                    "S16KQCX8XUO3EEL85N",
                    "OREKQCF34YE89RL8S6",
                    "RU5KQCQTPC9SJ5Q1JN",
                    "GYFKQCPH8Q0JDN94GR",
                    "9QFKQC54ET79K2SD57",
                    "MLEKQCH64557K610VR",
                    "AWFKQCJELLUWDY2PD3",
                    "ZDWJQC7TMS2AWAVF2Y",
                    "VJFKQCFO7WESWPNKHE",
                    "LV5KQCJCCR0S3DN5WW",
                    "CDTJQCVTVX7CNMY9YU",
                    "OX5KQCKE3I94MEPDC",
                    "JF5KQC2L6ABI0MVD3E",
                    "CH5KQCIEXSN1J5UEHB",
                    "4JEKQC22K5UTH9QHCU",
                    "EPFKQCETTDTEL3PYIR",
                    "D16KQCIZS56HVPW6DA",
                    "2TEKQCTSCRL4Z2AUHE",
                    "3WEKQCHXRH61E3CIKB",
                    "TDTJQCGYRS2RCCGQN3",
                    "SMWJQCXQ6I2GEXSPK9",
                    "PXFKQCXEHJX3W1Q1IV",
                    "Q9TJQCXDL1599L2B2Z",
                    "BFXJQCF1JBOXPRW2OS",
                    "6TFKQCUTO94WB2NHN",
                    "FNEKQCO239QSNK0R78",
                    "3RFKQCNKMX9HVLNSLW",
                    "W2DKQCV4H3EZUJ35DX",
                    "5IEKQCN37EFBK9EBUD",
                    "LW5KQCXSXRC2XV3T3D",
                    "ZNEKQCULV911DIXBK3",
                    "35WJQC1B7T7NPV8OLV",
                    "MPEKQC6TIP3SP1YF7I",
                    "DMFKQC5MHGYC6503F2",
                    "3KEKQC45RL87D4ZG86",
                    "KLTJQC70XVZJSPQ43U",
                    "LVEKQCP09W7JNFDAFC",
                    "DPUJQCUYKKDPT78JJV",
                    "JDXJQCR5L0NTR21SQA",
                    "UAXJQC6HC354V7Q6JA",
                    "XBXJQCS4QY316ZGHRN",
                    "HCXJQCRKB4K65V1C07",
                    "66TJQC6CC7ZWL9YZ64",
                    "XNFKQC6DN59LACS9IU",
                    "LL5KQCG687Y165GL5P",
                    "7X5KQC9ABK4T6AW7QF",
                    "HKTJQCIJD8RK9RJFO1",
                    "1LTJQCYQI1DXBLG6Z",
                    "0FWJQCLYEP736P3YZK",
                    "0DXJQC1T8P3CQKZIUO",
                    "ISEKQC97YI74A9VKWC",
                    "KGXJQCBQ39ON9VB37T",
                    "BZ5KQCD4KFI3BTIMZU",
                    "0HEKQCLINMQS4RB7B8",
                    "BBTJQCNT6N1H4X6TL4",
                    "COFKQCUXC2H4G3QBYT",
                    "DI5KQC3CLKWQI3I0CC",
                    "T8TJQCWWI8RY57YNTQ",
                    "46WJQCIOLQ0KOX2XCC",
                    "OEXJQCQJHF2BLSAAIS",
                    "GNFKQCH8AFJRJO9V4Y",
                    "8LFKQCQWXTJQJR4CXV",
                    "IGWJQCSVT8NXTFXOFJ",
                    "3PEKQC8ZDCYTSSIKZ9",
                    "5UEKQC8N8NEPSWU1QQ",
                    "NK5KQC1MAOU2BP35ZV",
                    "UZFKQCIHVT44DC9KGH",
                    "JJ5KQC83DT7VDMUYRQ",
                    "626KQC7C08H3UTM38E",
                    "XIFKQCLQOY645QTGP7",
                    "HEWJQCWQQ3K4BNRLIO",
                    "HDWJQCT361VOV2PQLP",
                    "XFKKQCGA4DVECEB48E",
                    "KWEKQCTNIIV9BROFFG",
                ]
            ),
        )

    def test_hasnickname(self):
        """
        Test HasNickname rule.
        """
        rule = HasNickname([])
        self.assertEqual(
            self.filter_with_rule(rule),
            set(
                [
                    "cc8205d883763f02abd",
                    "GNUJQCL9MD64AM56OH",
                    "Q8HKQC3VMRM1M6M7ES",
                ]
            ),
        )

    def test_hasunknowngender(self):
        """
        Test HasUnknownGender rule.
        """
        rule = HasUnknownGender([])
        self.assertEqual(
            self.filter_with_rule(rule),
            set(
                [
                    "OJOKQC83Y1EDBIMLJ6",
                    "8BHKQCFK9UZFRJYC2Y",
                    "PGFKQC1TUQMXFAMLMB",
                    "IHOKQCECRZYQDKW6KF",
                    "8HKKQCTEJAOBVH410L",
                    "AGFKQCO358R18LNJYV",
                    "1ENKQCBPFZTAQJSP4O",
                    "NUWKQCO7TVAOH0CHLV",
                    "P5IKQC88STY3FNTFZ3",
                    "7GXKQCMVFU8WR1LKZL",
                    "LGXKQCJ5OP6MKF9QLN",
                    "XNFKQC6DN59LACS9IU",
                    "7IOKQC1NVGUI1E55CQ",
                    "57PKQCFAWY7AM3JS4M",
                    "BNXKQCEBXC1RCOGJNF",
                    "TFFKQC1RMG8RRADKDH",
                    "FHKKQC963NGSY18ZDZ",
                    "WMXKQCDUJ4JKQQYCR7",
                    "PBHKQCHOAGTECRKT9L",
                    "OFXKQC8W0N3N6JP6YQ",
                ]
            ),
        )

    def test_hassourceof_empty(self):
        """
        Test empty HasSourceOf rule.
        """
        # when run with an empty string finds people with no sourc citations
        rule = HasSourceOf([""])
        self.assertEqual(
            self.filter_with_rule(rule),
            set(
                [
                    "cc82060512042f67e2c",
                    "cc8205d87831c772e87",
                    "cc82060516c6c141500",
                    "cc8205d87fd529000ff",
                    "cc82060504445ab6deb",
                    "cc8205d887376aacba2",
                    "cccbffffd3e69819cd8",
                    "cc8205d87c20350420b",
                    "cc8206050e541f79f92",
                    "cc8205d883763f02abd",
                    "cc8206050980ea622d0",
                    "cc8205d872f532ab14e",
                    "d5839c132b11d9e3632",
                    "d583a5ba0d50afbbaaf",
                    "d5839c1352c64b924d9",
                    "d583a5b9fc864e3bf4e",
                    "d583a5ba1bd083ce4c2",
                    "d583a5b9df71bceb48c",
                    "d583a5b9ced473a7e6a",
                    "d583a5ba2bc7b9d1388",
                    "d5839c12fec09785f6a",
                    "d5839c1237765987724",
                    "d5839c137b3640ad776",
                    "d5839c126d11a754f46",
                    "d5839c12d3b4d5e619b",
                    "d5839c13380462b246f",
                    "d5839c12e9e08301ce2",
                    "d5839c1366b21411fb4",
                    "d5839c13a282b51dd0d",
                    "d5839c12ac91650a72b",
                    "d583a5b9edf6cb5d8d5",
                    "d583a5ba4be3acdd312",
                    "d5839c131d560e06bac",
                    "d5839c13b0541b7b8e6",
                    "d5839c1388e3ab6c87c",
                    "d583a5ba5ca6b698463",
                    "d583a5ba3bc48c2002c",
                    "d583a5b90777391ea9a",
                ]
            ),
        )

    def test_hassourceof_nonmatching(self):
        """
        Test non-matching HasSourceOf rule.
        """
        rule = HasSourceOf(["S0004"])
        self.assertEqual(self.filter_with_rule(rule), set([]))

    def test_hassourceof_irregular(self):
        """
        Test irregular HasSourceOf rule.
        """
        rule = HasSourceOf(["ABCDEFG"])
        self.assertEqual(self.filter_with_rule(rule), set([]))

    def test_hassourceof_matching(self):
        """
        Test matching HasSourceOf rule.
        """
        rule = HasSourceOf(["S0000"])
        self.assertEqual(
            self.filter_with_rule(rule),
            set(
                [
                    "GNUJQCL9MD64AM56OH",
                ]
            ),
        )

    def test_havealtfamilies(self):
        """
        Test HaveAltFamilies rule.
        """
        rule = HaveAltFamilies([])
        self.assertEqual(
            self.filter_with_rule(rule),
            set(
                [
                    "CH5KQCIEXSN1J5UEHB",
                    "MG5KQC6ZKSVO4A63G2",
                ]
            ),
        )

    def test_havechildren(self):
        """
        Test HaveChildren rule.
        """
        rule = HaveChildren([])
        # too many to list out to test explicitly
        self.assertEqual(len(self.filter_with_rule(rule)), 905)

    def test_incompletenames(self):
        """
        Test IncompleteNames rule.
        """
        rule = IncompleteNames([])
        self.assertEqual(
            self.filter_with_rule(rule),
            set(
                [
                    "IHOKQCECRZYQDKW6KF",
                    "cc82060504445ab6deb",
                    "LCXKQCQZH5EH56NTCD",
                    "cc8205d87831c772e87",
                    "3RFKQCNKMX9HVLNSLW",
                    "cc8205d87fd529000ff",
                    "B1UKQCBR49WB3134PN",
                    "0TTKQCXXY59OCDPLV3",
                    "F3UKQC7ZV3EYVWTZ8O",
                    "1MXKQCJ2BR43910ZYX",
                    "cc8206050e541f79f92",
                    "FHKKQC963NGSY18ZDZ",
                    "R5HKQCIEPOY1DMQOWX",
                    "ZHMKQC50PFVAPI8PZ6",
                    "T4UKQCYGECXGVNBWMY",
                    "cc82060516c6c141500",
                    "UPWKQCYVFH7RZOSZ29",
                    "2AMKQCE67YOH3TBVYI",
                    "2CUKQCFDVN3EZE2E4C",
                    "7IOKQC1NVGUI1E55CQ",
                    "KSTKQC018GNA7HDCAS",
                    "WIVKQC4Q4FCQJT5M63",
                    "A4YKQCRYSI5FT5T38",
                    "BUNKQCO4HZHZP70F3K",
                    "YRTKQCNDP343OD5OQJ",
                    "7VEKQCV05EDK0625KI",
                    "cc8205d872f532ab14e",
                    "TPXKQCEGL04KHGMO2X",
                    "L9LKQCQ8KJRKHM4D2E",
                    "8QXKQCHJ2EUC7OV8EQ",
                    "W0XKQCKSFWWJWQ2OSN",
                    "I6QKQCFRDTV2LDC8M2",
                    "XTUKQC7WCIVA5F0NC4",
                    "F4UKQCPK572VWU2YZQ",
                    "JKDKQCF4ND92A088J2",
                    "COFKQCUXC2H4G3QBYT",
                    "BNXKQCEBXC1RCOGJNF",
                    "Q42KQCKJZGS4IZWHF5",
                    "P5IKQC88STY3FNTFZ3",
                    "7CXKQC59NSZFXIG1UE",
                    "cc8205d87c20350420b",
                    "FQUKQCWEHOAWUP4QWS",
                    "3YTKQCK2W63W0MQBJE",
                    "8HKKQCTEJAOBVH410L",
                    "HLQKQC0BJIZL0V4EK4",
                    "B0UKQC9A54F1GUB7NR",
                    "EPXKQCQRZP2PNPN7BE",
                    "cc82060512042f67e2c",
                    "XZLKQCRQA9EHPBNZPT",
                    "OQXKQC2Y5FVH9PK0JL",
                    "AXLKQC0YTFAWQ234YD",
                    "OFXKQC8W0N3N6JP6YQ",
                    "MWUKQCD2ZSCECQOCLG",
                    "1ENKQCBPFZTAQJSP4O",
                    "N7XKQCYD3VSCSZREGJ",
                    "2LQKQC62GJUQCJIOK8",
                    "QXXKQC9PT5FWNT140K",
                    "VAXKQC19HIFPX61J28",
                    "0PXKQCJ9S1M3NNASET",
                    "K8XKQCDSVLSK422A3K",
                    "52UKQCFYXMFTKIGNBS",
                    "7GXKQCMVFU8WR1LKZL",
                    "4UMKQCF07KL2K92CI5",
                    "LGXKQCJ5OP6MKF9QLN",
                    "FZTKQCSTPIQ3C9JC46",
                    "WMXKQCDUJ4JKQQYCR7",
                    "R6UKQC939L9FV62UGE",
                    "OIUKQCBHUWDGL7DNTI",
                    "FRTKQC3G6JBJAR2ZPX",
                    "PIEKQCKUL6OAMS8Q9R",
                    "cc8205d887376aacba2",
                    "LGMKQCQP5M5L18FVTN",
                    "8HUKQCRV8B3J2LLQ3B",
                    "LOUKQC45HUN532HOOM",
                    "cc8205d883763f02abd",
                    "TBXKQC7OHIN28PVCS3",
                ]
            ),
        )

    def test_isbookmarked(self):
        """
        Test IsBookmarked rule.
        """
        rule = IsBookmarked([])
        self.assertEqual(
            self.filter_with_rule(rule),
            set(
                [
                    "35WJQC1B7T7NPV8OLV",
                    "AWFKQCJELLUWDY2PD3",
                    "Q8HKQC3VMRM1M6M7ES",
                ]
            ),
        )

    def test_dupancestor_empty(self):
        """
        Test empty IsDuplicatedAncestorOf rule.
        """
        rule = IsDuplicatedAncestorOf([""])
        self.assertEqual(self.filter_with_rule(rule), set([]))

    def test_dupancestor_nonmatching(self):
        """
        Test non-matching IsDuplicatedAncestorOf rule.
        """
        rule = IsDuplicatedAncestorOf(["I0000"])
        self.assertEqual(self.filter_with_rule(rule), set([]))

    def test_dupancestor_irregular(self):
        """
        Test irregular IsDuplicatedAncestorOf rule.
        """
        rule = IsDuplicatedAncestorOf(["ABCDEFG"])
        self.assertEqual(self.filter_with_rule(rule), set([]))

    def test_dupancestor_matching(self):
        """
        Test matching IsDuplicatedAncestorOf rule.
        """
        rule = IsDuplicatedAncestorOf(["I1631"])
        self.assertEqual(
            self.filter_with_rule(rule),
            set(
                [
                    "I3VJQCUY5I6UR92507",
                    "D4VJQC09STQCWD393E",
                ]
            ),
        )

    def test_isrelatedwith_empty(self):
        """
        Test empty IsRelatedWith rule.
        """
        rule = IsRelatedWith([""])
        self.assertEqual(self.filter_with_rule(rule), set([]))

    def test_isrelatedwith_nonmatching(self):
        """
        Test non-matching IsRelatedWith rule.
        """
        rule = IsRelatedWith(["I0000"])
        self.assertEqual(
            self.filter_with_rule(rule),
            set(
                [
                    "d5839c1237765987724",
                    "d5839c126d11a754f46",
                ]
            ),
        )

    def test_isrelatedwith_irregular(self):
        """
        Test irregular IsRelatedWith rule.
        """
        rule = IsRelatedWith(["ABCDEFG"])
        self.assertEqual(self.filter_with_rule(rule), set([]))

    def test_isrelatedwith_matching(self):
        """
        Test matching IsRelatedWith rule.
        """
        rule = IsRelatedWith(["I1844"])
        self.assertEqual(
            self.filter_with_rule(rule),
            set(
                [
                    "HWTKQCSM28EI6WFDHP",
                    "T4UKQCYGECXGVNBWMY",
                    "YOTKQCEX2PLG03LZQS",
                    "X8UKQCIDY21QIQBDVI",
                    "F3UKQC7ZV3EYVWTZ8O",
                    "0TTKQCXXY59OCDPLV3",
                    "EVTKQCHV2E2PODFD7C",
                    "BBUKQC5GPRPDJHJAWU",
                    "FRTKQC3G6JBJAR2ZPX",
                    "NDTKQCN95VFLGJ21L",
                    "SFTKQC26EJ2BYQCRIA",
                    "MYTKQCVCFOFM32H9GB",
                    "B0UKQC9A54F1GUB7NR",
                    "PTTKQCYN0JR3ZZJNWR",
                    "F4UKQCPK572VWU2YZQ",
                    "LLTKQCX39KCXFSX0U4",
                    "IXTKQC1BAU1F1WNXKB",
                    "3YTKQCK2W63W0MQBJE",
                    "TQTKQCO897BNA1H93B",
                    "DOTKQCP1MG3VC8D7V2",
                    "3NTKQCZKLMIM6HYFE1",
                    "WUTKQCVQCUPFFOGUT8",
                    "GETKQCPRC2W5YDUYM6",
                    "YRTKQCNDP343OD5OQJ",
                    "U0UKQCBZS0R6WW7LBS",
                    "J2UKQC897I42M9VHDD",
                    "7MTKQC1QNE4H5RF35S",
                    "5FTKQCKT9SDZ8TB03C",
                    "O1UKQCJD5YHDRW887V",
                    "EUTKQCFATXRU431YY6",
                    "UHTKQCORH3NTZ0FYL3",
                    "2CUKQCFDVN3EZE2E4C",
                    "RNTKQCMLGRRKQVKDPR",
                    "CGTKQC4WO8W3WSQRCX",
                    "WAUKQCOQ91QCJZWQ9U",
                    "FZTKQCSTPIQ3C9JC46",
                    "AHTKQCM2YFRW3AGSRL",
                    "WBTKQCC775IAAGIWZD",
                    "8KTKQC407A8CN5O68H",
                    "8QTKQCN8ZKY5OWWJZF",
                    "UKTKQCSL3AUJIWTD2A",
                    "HAUKQCM3GYGVTREGZS",
                    "52UKQCFYXMFTKIGNBS",
                    "U3UKQCO30PWAK6JQBA",
                    "R6UKQC939L9FV62UGE",
                    "TZTKQCR39A060AQ63C",
                    "X9UKQCFELSDAQ2TDP1",
                    "B1UKQCBR49WB3134PN",
                    "KSTKQC018GNA7HDCAS",
                    "FJTKQCJCMAHJOA9NHI",
                    "HITKQCWJSCZX2AN6NP",
                    "WVTKQCZC91I63LHEE7",
                    "0DTKQC6KBOS69LQJ35",
                ]
            ),
        )

    def test_hasidof_empty(self):
        """
        Test empty HasIdOf rule.
        """
        rule = HasIdOf([""])
        self.assertEqual(self.filter_with_rule(rule), set([]))

    def test_hasidof_nonmatching(self):
        """
        Test non-matching HasIdOf rule.
        """
        rule = HasIdOf(["I0000"])
        self.assertEqual(self.filter_with_rule(rule), set(["d5839c1237765987724"]))

    def test_hasidof_irregular(self):
        """
        Test irregular HasIdOf rule.
        """
        rule = HasIdOf(["ABCDEFG"])
        self.assertEqual(self.filter_with_rule(rule), set([]))

    def test_hasidof_matching(self):
        """
        Test matching HasIdOf rule.
        """
        rule = HasIdOf(["I0044"])
        self.assertEqual(
            self.filter_with_rule(rule),
            set(
                [
                    "GNUJQCL9MD64AM56OH",
                ]
            ),
        )

    def test_isdefaultperson(self):
        """
        Test IsDefaultPerson rule.
        """
        rule = IsDefaultPerson([])
        self.assertEqual(
            self.filter_with_rule(rule),
            set(
                [
                    "GNUJQCL9MD64AM56OH",
                ]
            ),
        )

    def test_isdefaultperson_optimized(self):
        """
        Test IsDefaultPerson rule.
        """
        with count_method_calls(IsDefaultPerson, "apply_to_one") as get_call_count:
            rule = IsDefaultPerson([])
            self.assertEqual(
                self.filter_with_rule(rule),
                set(
                    [
                        "GNUJQCL9MD64AM56OH",
                    ]
                ),
            )
            # This used the optimizer, so it didn't loop through DB
            self.assertEqual(get_call_count(), 1)

    def test_isfemale_not_optimized(self):
        """
        Test IsFemale rule. Same as below, but tests optimizer.
        """
        with count_method_calls(IsFemale, "apply_to_one") as get_call_count:
            rule = IsFemale([])
            # too many to list out to test explicitly
            self.assertEqual(len(self.filter_with_rule(rule)), 940)
            # This did not use the optimizer, so it did loop through DB
            self.assertEqual(get_call_count(), self.db.get_number_of_people())

    def test_isfemale(self):
        """
        Test IsFemale rule.
        """
        rule = IsFemale([])
        # too many to list out to test explicitly
        self.assertEqual(len(self.filter_with_rule(rule)), 940)

    def test_ismale(self):
        """
        Test IsMale rule.
        """
        rule = IsMale([])
        # too many to list out to test explicitly
        self.assertEqual(len(self.filter_with_rule(rule)), 1168)

    def test_missingparent(self):
        """
        Test MissingParent rule.
        """
        rule = MissingParent([])
        # too many to list out to test explicitly
        self.assertEqual(len(self.filter_with_rule(rule)), 868)

    def test_multiplemarriages(self):
        """
        Test MultipleMarriages rule.
        """
        rule = MultipleMarriages([])
        self.assertEqual(
            self.filter_with_rule(rule),
            set(
                [
                    "R1VKQCJWNP24VN7BO",
                    "ZTVJQCTSMI85EGMXFM",
                    "ENTJQCZXQV1IRKJXUL",
                    "44WJQCLCQIPZUB0UH",
                    "SMWJQCXQ6I2GEXSPK9",
                    "DN3KQC1URTED410L3R",
                    "5FYJQC86G8EZ0L4E4B",
                    "5F4KQCJRU8ZKL6SILT",
                    "0YNKQC5U4EQGVNUZD8",
                    "YRYJQCE3RF4U8A59UB",
                    "APWKQCI6YXAXBLC33I",
                    "XSKKQC6GGKLAYANWAF",
                    "0FQKQCOQD0VRVJPTSD",
                    "B3UJQCZHDXII99AWW4",
                    "cc8205d872f532ab14e",
                    "SS1KQCWWF9488Q330U",
                    "OCYJQCS8YT7JO8KIMO",
                    "I6HKQCQF72V2N56JQ5",
                    "6YWJQC86FBVN0J6JS",
                    "KYNKQCVA6FE65ONFIQ",
                    "SHAKQCNY5IXO30GUAB",
                    "O5XKQC3V6BPJI13J24",
                    "ZN7KQC3RLB82EXF1QF",
                    "CIYJQCF3UK12DL0S2Y",
                    "H3XJQCFJ4FP4U2WGZC",
                    "cc82060504445ab6deb",
                    "4E4KQC1K4XUEX29IJO",
                    "0XVJQCJUNJY40WDSMA",
                    "1WUJQCHNH76G6YD3A",
                    "IH3KQCM1VZPRKLBLK7",
                    "242KQCBALBOD8ZK5VI",
                    "8G4KQCS6C1AOM6ZGR3",
                    "I1EKQCGGDSUD8ILUW4",
                    "X8BKQCSFF4AET5MY23",
                    "RJWJQCN1XKXRN5KMCP",
                    "ZWNKQC9DAZ3C6UHUAV",
                    "9QUJQCCSWRZNSAPCR",
                    "HI0KQCG9TGT5AAIPU",
                    "DI4KQC3S1AO27VWOLN",
                    "QBDKQCH2IU6N8IXMFE",
                    "DK2KQCJYW14VXUJ85",
                    "117KQCBB32RMTTV4G6",
                    "0QLKQCFTQMNVGCV4GM",
                    "D2OKQCGDNPT3BH4WH",
                    "CAYJQCKOL49OF7XWB3",
                    "ZQGKQCGHS67Q4IMHEG",
                    "OEXJQCQJHF2BLSAAIS",
                    "UKYJQC70LIZQ11BP89",
                    "FF2KQCRBSPCG1QY97",
                    "L6EKQCO8QYL2UO2MQO",
                ]
            ),
        )

    def test_nevermarried(self):
        """
        Test NeverMarried rule.
        """
        rule = NeverMarried([])
        # too many to list out to test explicitly
        self.assertEqual(len(self.filter_with_rule(rule)), 751)

    def test_nobirthdate(self):
        """
        Test NoBirthdate rule.
        """
        rule = NoBirthdate([])
        # too many to list out to test explicitly
        self.assertEqual(len(self.filter_with_rule(rule)), 981)

    def test_nodeathdate(self):
        """
        Test NoDeathdate rule.
        """
        rule = NoDeathdate([])
        # too many to list out to test explicitly
        self.assertEqual(len(self.filter_with_rule(rule)), 1603)

    def test_peopleprivate(self):
        """
        Test PeoplePrivate rule.
        """
        rule = PeoplePrivate([])
        self.assertEqual(self.filter_with_rule(rule), set(["0GDKQC54XKSWZKEBWW"]))

    def test_peoplepublic(self):
        """
        Test PeoplePublic rule.
        """
        rule = PeoplePublic([])
        # too many to list out to test explicitly
        self.assertEqual(len(self.filter_with_rule(rule)), 2127)

    def test_personwithincompleteevent(self):
        """
        Test PersonWithIncompleteEvent rule.
        """
        rule = PersonWithIncompleteEvent([])
        # too many to list out to test explicitly
        self.assertEqual(len(self.filter_with_rule(rule)), 745)

    def test_relationshipbookmarks(self):
        """
        Test RelationshipPathBetweenBookmarks rule.
        """
        rule = RelationshipPathBetweenBookmarks([])
        self.assertEqual(
            self.filter_with_rule(rule),
            set(
                [
                    "44WJQCLCQIPZUB0UH",
                    "35WJQC1B7T7NPV8OLV",
                    "AWFKQCJELLUWDY2PD3",
                    "D3WJQCCGV58IP8PNHZ",
                    "Q8HKQC3VMRM1M6M7ES",
                ]
            ),
        )

    def test_hassoundexname(self):
        """
        Test HasSoundexName rule.
        """
        rule = HasSoundexName(["garner"])
        self.assertEqual(len(self.filter_with_rule(rule)), 73)

    def test_hasnameof(self):
        """
        Test HasNameOf rule.
        """
        rule = HasNameOf(
            [
                "Lewis",
                "Garner",
                "Dr.",
                "Sr",
                "Anderson",
                "Big Louie",
                "von",
                "Zieliński",
                None,
                None,
                None,
            ]
        )
        self.assertEqual(self.filter_with_rule(rule), set(["GNUJQCL9MD64AM56OH"]))


if __name__ == "__main__":
    unittest.main()



#
# Gramps - a GTK+/GNOME based genealogy program
#
# Copyright (C) 2018        Nick Hall
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#

__all__ = ["ShadeBox"]

# -------------------------------------------------------------------------
#
# Standard python modules
#
# -------------------------------------------------------------------------
import logging

_LOG = logging.getLogger(".widgets.shadebox")

# -------------------------------------------------------------------------
#
# GTK/Gnome modules
#
# -------------------------------------------------------------------------
from gi.repository import Gtk


# -------------------------------------------------------------------------
#
# ShadeBox class
#
# -------------------------------------------------------------------------
class ShadeBox(Gtk.EventBox):
    """
    An EventBox with a shaded background.
    """

    def __init__(self, use_shade):
        Gtk.EventBox.__init__(self)
        self.use_shade = use_shade

    def do_draw(self, cr):
        if self.use_shade:
            tv = Gtk.TextView()
            tv_context = tv.get_style_context()
            width = self.get_allocated_width()
            height = self.get_allocated_height()
            Gtk.render_background(tv_context, cr, 0, 0, width, height)
        self.get_child().draw(cr)



#
# Gramps - a GTK+/GNOME based genealogy program
#
# Copyright (C) 2008  Zsolt Foldvari
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#

"""
Define text formatting tag types.
"""

# -------------------------------------------------------------------------
#
# Gramps modules
#
# -------------------------------------------------------------------------
from ..const import GRAMPS_LOCALE as glocale
from .grampstype import GrampsType

_ = glocale.translation.gettext


# -------------------------------------------------------------------------
#
# StyledTextTagType
#
# -------------------------------------------------------------------------
class StyledTextTagType(GrampsType):
    """
    Text formatting tag type definition.

    Here we only define new class variables. For details see
    :class:`~gen.lib.grampstype.GrampsType`.
    """

    NONE_TYPE = -1
    BOLD = 0
    ITALIC = 1
    UNDERLINE = 2
    FONTFACE = 3
    FONTSIZE = 4
    FONTCOLOR = 5
    HIGHLIGHT = 6
    SUPERSCRIPT = 7
    LINK = 8
    STRIKETHROUGH = 9
    SUBSCRIPT = 10

    _CUSTOM = NONE_TYPE
    _DEFAULT = NONE_TYPE

    _DATAMAP = [
        (BOLD, _("Bold"), "bold"),
        (ITALIC, _("Italic"), "italic"),
        (UNDERLINE, _("Underline"), "underline"),
        (FONTFACE, _("Fontface"), "fontface"),
        (FONTSIZE, _("Fontsize"), "fontsize"),
        (FONTCOLOR, _("Fontcolor"), "fontcolor"),
        (HIGHLIGHT, _("Highlight"), "highlight"),
        (SUPERSCRIPT, _("Superscript"), "superscript"),
        (LINK, _("Link"), "link"),
        (STRIKETHROUGH, _("Strikethrough"), "strikethrough"),
        (SUBSCRIPT, _("Subscript"), "subscript"),
    ]

    STYLE_TYPE = {
        BOLD: bool,
        ITALIC: bool,
        UNDERLINE: bool,
        FONTCOLOR: str,
        HIGHLIGHT: str,
        FONTFACE: str,
        FONTSIZE: int,
        SUPERSCRIPT: bool,
        LINK: str,
        STRIKETHROUGH: bool,
        SUBSCRIPT: bool,
    }

    STYLE_DEFAULT = {
        BOLD: False,
        ITALIC: False,
        UNDERLINE: False,
        FONTCOLOR: "#000000",
        HIGHLIGHT: "#FFFFFF",
        FONTFACE: "Sans",
        FONTSIZE: 10,
        SUPERSCRIPT: False,
        LINK: "",
        STRIKETHROUGH: False,
        SUBSCRIPT: False,
    }

    def __init__(self, value=None):
        GrampsType.__init__(self, value)



#
# Gramps - a GTK+/GNOME based genealogy program
#
# Copyright (C) 2016 Tom Samstag
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#

"""
Unittest that tests family-specific filter rules
"""
import unittest
import os

from ....db.utils import import_as_dict
from ....filters import GenericFilterFactory
from ....const import DATA_DIR
from ....user import User
from ....utils.unittest import localize_date

from ..family import (
    AllFamilies,
    HasRelType,
    HasGallery,
    HasIdOf,
    HasLDS,
    HasNote,
    RegExpIdOf,
    HasNoteRegexp,
    HasReferenceCountOf,
    HasSourceCount,
    HasSourceOf,
    HasCitation,
    FamilyPrivate,
    HasEvent,
    HasAttribute,
    IsBookmarked,
    MatchesSourceConfidence,
    FatherHasNameOf,
    FatherHasIdOf,
    MotherHasNameOf,
    MotherHasIdOf,
    ChildHasNameOf,
    ChildHasIdOf,
    ChangedSince,
    HasTag,
    HasTwins,
    IsAncestorOf,
    IsDescendantOf,
)

TEST_DIR = os.path.abspath(os.path.join(DATA_DIR, "tests"))
EXAMPLE = os.path.join(TEST_DIR, "example.gramps")
GenericFamilyFilter = GenericFilterFactory("Family")


class BaseTest(unittest.TestCase):
    """
    Family rule tests.
    """

    @classmethod
    def setUpClass(cls):
        """
        Import example database.
        """
        cls.db = import_as_dict(EXAMPLE, User())

    def filter_with_rule(self, rule):
        """
        Apply a filter with the given rule.
        """
        filter_ = GenericFamilyFilter()
        filter_.add_rule(rule)
        results = filter_.apply(self.db)
        return set(results)

    def test_allfamilies(self):
        """
        Test AllFamilies rule.
        """
        rule = AllFamilies([])
        self.assertEqual(
            len(self.filter_with_rule(rule)), self.db.get_number_of_families()
        )

    def test_hasreltype(self):
        """
        Test HasRelType rule.
        """
        rule = HasRelType(["Married"])
        self.assertEqual(len(self.filter_with_rule(rule)), 750)

    def test_hasgallery(self):
        """
        Test HasGallery rule.
        """
        rule = HasGallery(["0", "greater than"])
        self.assertEqual(self.filter_with_rule(rule), set(["9OUJQCBOHW9UEK9CNV"]))

    def test_hasidof(self):
        """
        Test HasIdOf rule.
        """
        rule = HasIdOf(["F0001"])
        self.assertEqual(self.filter_with_rule(rule), set(["48TJQCGNNIR5SJRCAK"]))

    def test_haslds(self):
        """
        Test HasLDS rule.
        """
        rule = HasLDS(["0", "greater than"])
        self.assertEqual(self.filter_with_rule(rule), set(["9OUJQCBOHW9UEK9CNV"]))

    def test_hasnote(self):
        """
        Test HasNote rule.
        """
        rule = HasNote([])
        self.assertEqual(self.filter_with_rule(rule), set(["9OUJQCBOHW9UEK9CNV"]))

    def test_regexpidof(self):
        """
        Test RegExpIdOf rule.
        """
        rule = RegExpIdOf(["F000."], use_regex=True)
        self.assertEqual(
            self.filter_with_rule(rule),
            set(
                [
                    "LOTJQC78O5B4WQGJRP",
                    "UPTJQC4VPCABZUDB75",
                    "NBTJQCIX49EKOCIHBP",
                    "C9UJQCF6ETBTV2MRRV",
                    "74UJQCKV8R4NBNHCB",
                    "4BTJQCL4CHNA5OUTKF",
                    "48TJQCGNNIR5SJRCAK",
                    "4YTJQCTEH7PQUU4AD",
                    "MTTJQC05LKVFFLN01A",
                    "d5839c123c034ef82ab",
                ]
            ),
        )

    def test_hasnoteregexp(self):
        """
        Test HasNoteRegexp rule.
        """
        rule = HasNoteRegexp(["."], use_regex=True)
        self.assertEqual(self.filter_with_rule(rule), set(["9OUJQCBOHW9UEK9CNV"]))

    def test_hasreferencecountof(self):
        """
        Test HasReferenceCountOf rule.
        """
        rule = HasReferenceCountOf(["greater than", "12"])
        self.assertEqual(
            self.filter_with_rule(rule),
            set(
                [
                    "29IKQCMUNFTIBV653N",
                    "8OUJQCUVZ0XML7BQLF",
                    "UPTJQC4VPCABZUDB75",
                    "9NWJQCJGLXUR3AQSFJ",
                    "5G2KQCGBTS86UVSRG5",
                    "WG2KQCSY9LEFDFQHMN",
                    "MTTJQC05LKVFFLN01A",
                    "C2VJQC71TNHO7RBBMX",
                    "QIDKQCJQ37SIUQ3UFU",
                    "DV4KQCX9OBVQ74H77F",
                ]
            ),
        )

    def test_hassourcecount(self):
        """
        Test HasSourceCount rule.
        """
        rule = HasSourceCount(["1", "greater than"])
        self.assertEqual(self.filter_with_rule(rule), set(["9OUJQCBOHW9UEK9CNV"]))

    def test_hassourceof(self):
        """
        Test HasSourceOf rule.
        """
        rule = HasSourceOf(["S0001"])
        self.assertEqual(self.filter_with_rule(rule), set(["9OUJQCBOHW9UEK9CNV"]))

    def test_hascitation(self):
        """
        Test HasCitation rule.
        """
        rule = HasCitation(["page 10", "", "2"])
        self.assertEqual(self.filter_with_rule(rule), set(["9OUJQCBOHW9UEK9CNV"]))

    def test_familyprivate(self):
        """
        Test FamilyPrivate rule.
        """
        rule = FamilyPrivate([])
        self.assertEqual(self.filter_with_rule(rule), set([]))

    def test_hasevent(self):
        """
        Test HasEvent rule.
        """
        date_str = localize_date("before 1900")
        rule = HasEvent(["Marriage", date_str, "USA", "", "Garner"])
        self.assertEqual(
            self.filter_with_rule(rule),
            set(
                [
                    "KSFKQCP4V0YXGM1LR9",
                    "8ZFKQC3FRSHACOJBOU",
                    "3XFKQCE7QUDJ99AVNV",
                    "OVFKQC51DX0OQUV3JB",
                    "9OUJQCBOHW9UEK9CNV",
                ]
            ),
        )

    def test_hasattribute(self):
        """
        Test HasAttribute rule.
        """
        rule = HasAttribute(["Number of Children", ""])
        self.assertEqual(self.filter_with_rule(rule), set(["9OUJQCBOHW9UEK9CNV"]))

    def test_isbookmarked(self):
        """
        Test IsBookmarked rule.
        """
        rule = IsBookmarked([])
        self.assertEqual(self.filter_with_rule(rule), set(["9OUJQCBOHW9UEK9CNV"]))

    def test_matchessourceconfidence(self):
        """
        Test MatchesSourceConfidence rule.
        """
        rule = MatchesSourceConfidence(["0"])
        self.assertEqual(len(self.filter_with_rule(rule)), 734)

    def test_fatherhasnameof(self):
        """
        Test FatherHasNameOf rule.
        """
        rule = FatherHasNameOf(["", "", "Dr.", "", "", "", "", "", "", "", ""])
        self.assertEqual(self.filter_with_rule(rule), set(["9OUJQCBOHW9UEK9CNV"]))

    def test_fatherhasidof(self):
        """
        Test FatherHasIdOf rule.
        """
        rule = FatherHasIdOf(["I0106"])
        self.assertEqual(self.filter_with_rule(rule), set(["8OUJQCUVZ0XML7BQLF"]))

    def test_motherhasnameof(self):
        """
        Test MotherHasNameOf rule.
        """
        rule = MotherHasNameOf(["", "Alvarado", "", "", "", "", "", "", "", "", ""])
        self.assertEqual(
            self.filter_with_rule(rule),
            set(
                [
                    "EM3KQC48HFLA02TF8D",
                    "K9NKQCBG105ECXZ48D",
                    "2QMKQC5YWNAWZMG6VO",
                    "6JUJQCCAXGENRX990K",
                ]
            ),
        )

    def test_motherhasidof(self):
        """
        Test MotherHasIdOf rule.
        """
        rule = MotherHasIdOf(["I0107"])
        self.assertEqual(self.filter_with_rule(rule), set(["8OUJQCUVZ0XML7BQLF"]))

    def test_childhasnameof(self):
        """
        Test ChildHasNameOf rule.
        """
        rule = ChildHasNameOf(["Eugene", "", "", "", "", "", "", "", "", "", ""])
        self.assertEqual(
            self.filter_with_rule(rule),
            set(
                [
                    "D1YJQCGLEIBPPLNL4B",
                    "5GTJQCXVYVAIQTBVKA",
                    "I42KQCM3S926FMJ91O",
                    "7CTJQCFJVBQSY076A6",
                    "9OUJQCBOHW9UEK9CNV",
                    "9IXJQCX18AHUFPQHEZ",
                    "9NWJQCJGLXUR3AQSFJ",
                ]
            ),
        )

    def test_childhasidof(self):
        """
        Test ChildHasIdOf rule.
        """
        rule = ChildHasIdOf(["I0001"])
        self.assertEqual(self.filter_with_rule(rule), set(["48TJQCGNNIR5SJRCAK"]))

    def test_changedsince(self):
        """
        Test ChangedSince rule.
        """
        rule = ChangedSince(["2008-01-01", "2014-01-01"])
        self.assertEqual(self.filter_with_rule(rule), set(["9OUJQCBOHW9UEK9CNV"]))

    def test_hastag(self):
        """
        Test HasTag rule.
        """
        rule = HasTag(["ToDo"])
        self.assertEqual(self.filter_with_rule(rule), set(["9OUJQCBOHW9UEK9CNV"]))

    def test_hastwins(self):
        """
        Test HasTwins rule.
        """
        rule = HasTwins([])
        self.assertEqual(
            self.filter_with_rule(rule),
            set(
                [
                    "SD6KQC7LB8MYGA7F5W",
                    "8OUJQCUVZ0XML7BQLF",
                    "1BVJQCNTFAGS8273LJ",
                    "5IUJQCRJY47YQ8PU7N",
                    "ZLUJQCPDV93OR8KHB7",
                    "4U2KQCBXG2VTPH6U1F",
                ]
            ),
        )

    def test_isancestorof(self):
        """
        Test IsAncestorOf rule.
        """
        rule = IsAncestorOf(["F0031", "0"])
        self.assertEqual(
            self.filter_with_rule(rule),
            set(
                [
                    "4AXJQC96KTN3WGPTVE",
                    "1RUJQCYX9QL1V45YLD",
                    "5GTJQCXVYVAIQTBVKA",
                    "X3WJQCSF48F6809142",
                    "NSVJQC89IHEEBIPDP2",
                    "9OUJQCBOHW9UEK9CNV",
                    "1RUJQCCL9MVRYLMTBO",
                    "RRVJQC5A8DDHQFPRDL",
                    "0SUJQCOS78AXGWP8QR",
                    "57WJQCTBJKR5QYPS6K",
                    "8OUJQCUVZ0XML7BQLF",
                    "7PUJQC4PPS4EDIVMYE",
                ]
            ),
        )

    def test_isdescendantof(self):
        """
        Test IsDescendantOf rule.
        """
        rule = IsDescendantOf(["F0031", "0"])
        self.assertEqual(
            self.filter_with_rule(rule),
            set(
                [
                    "SFXJQCLE8PIG7PH38J",
                    "UCXJQCC5HS8VXDKWBM",
                    "IIEKQCRX89WYBHKB7R",
                    "XDXJQCMWU5EIV8XCRF",
                    "7BXJQCU22OCA4HN38A",
                    "3FXJQCR749H2H7G321",
                    "IEXJQCFUN95VENI6BO",
                    "4FXJQC7656WDQ3HJGW",
                    "FLEKQCRVG3O1UA9YUB",
                    "BCXJQC9AQ0DBXCVLEQ",
                    "9SEKQCAAWRUCIO7A0M",
                    "DDXJQCVT5X72TOXP0C",
                    "CGXJQC515QL9RLPQTU",
                    "XGXJQCNVZH2PWRMVAH",
                    "RBXJQCUYMQR2KRMDFY",
                ]
            ),
        )


if __name__ == "__main__":
    unittest.main()



#
# Gramps - a GTK+/GNOME based genealogy program
#
# Copyright (C) 2000-2006  Donald N. Allingham
# Copyright (C) 2011       Tim G L Lyons
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#


# -------------------------------------------------------------------------
#
# set up logging
#
# -------------------------------------------------------------------------
import logging

log = logging.getLogger("gui.editors.displaytabs")

# first import models
from .childmodel import ChildModel

# Then import tab classes
from .grampstab import GrampsTab
from .embeddedlist import EmbeddedList, TEXT_COL, MARKUP_COL, ICON_COL
from .addrembedlist import AddrEmbedList
from .attrembedlist import AttrEmbedList
from .backreflist import BackRefList
from .eventattrembedlist import EventAttrEmbedList
from .eventbackreflist import EventBackRefList
from .eventembedlist import EventEmbedList
from .familyattrembedlist import FamilyAttrEmbedList
from .familyldsembedlist import FamilyLdsEmbedList
from .gallerytab import GalleryTab
from .ldsembedlist import LdsEmbedList
from .locationembedlist import LocationEmbedList
from .mediaattrembedlist import MediaAttrEmbedList
from .mediabackreflist import MediaBackRefList
from .nameembedlist import NameEmbedList
from .notebackreflist import NoteBackRefList
from .notetab import NoteTab
from .citationbackreflist import CitationBackRefList
from .citationembedlist import CitationEmbedList
from .personeventembedlist import PersonEventEmbedList
from .personrefembedlist import PersonRefEmbedList
from .personbackreflist import PersonBackRefList
from .placebackreflist import PlaceBackRefList
from .placenameembedlist import PlaceNameEmbedList
from .placerefembedlist import PlaceRefEmbedList
from .repoembedlist import RepoEmbedList
from .surnametab import SurnameTab
from .sourcebackreflist import SourceBackRefList
from .srcattrembedlist import SrcAttrEmbedList
from .webembedlist import WebEmbedList



#
# Gramps - a GTK+/GNOME based genealogy program
#
# Copyright (C) 2000-2006  Donald N. Allingham
#               2009       Benny Malengier
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#

# -------------------------------------------------------------------------
#
# python
#
# -------------------------------------------------------------------------
import pickle

# -------------------------------------------------------------------------
#
# GTK libraries
#
# -------------------------------------------------------------------------
from gi.repository import GObject
from gi.repository import GLib
from gi.repository import Gdk
from gi.repository import Gtk
from gi.repository import Pango

# -------------------------------------------------------------------------
#
# Gramps classes
#
# -------------------------------------------------------------------------
from ...widgets.cellrenderertextedit import CellRendererTextEdit
from ...widgets.persistenttreeview import PersistentTreeView
from gramps.gen.const import GRAMPS_LOCALE as glocale

_ = glocale.translation.gettext
from ...utils import is_right_click
from .buttontab import ButtonTab


# ----------------------------------------------------------------
#
# Constants
#
# ----------------------------------------------------------------
TEXT_COL = 0
MARKUP_COL = 1
ICON_COL = 2
TEXT_EDIT_COL = 3


# -------------------------------------------------------------------------
#
# Classes
#
# -------------------------------------------------------------------------
class EmbeddedList(ButtonTab):
    """
    This class provides the base class for all the list tabs.

    It maintains a Gtk.TreeView, including the selection and button sensitivity.
    """

    _HANDLE_COL = -1
    _DND_TYPE = None
    _DND_EXTRA = None

    def __init__(
        self,
        dbstate,
        uistate,
        track,
        name,
        build_model,
        config_key,
        share_button=False,
        move_buttons=False,
        jump_button=False,
        top_label=None,
    ):
        """
        Create a new list, using the passed build_model to populate the list.
        """
        self.config_key = config_key

        ButtonTab.__init__(
            self,
            dbstate,
            uistate,
            track,
            name,
            share_button,
            move_buttons,
            jump_button,
            top_label,
        )

        self.changed = False
        self.model = None
        self.build_model = build_model
        # renderer for pixbuf
        self.pb_renderer = None

        # handle the selection
        self.selection = self.tree.get_selection()
        self.selection.connect("changed", self._selection_changed)
        self.track_ref_for_deletion("selection")

        # build the columns
        self.col_icons = {}
        self.columns = []
        self.build_columns()

        if self._DND_TYPE:
            self._set_dnd()

        # set up right click option
        self.tree.connect("button-press-event", self._on_button_press)

        # build the initial data
        self.rebuild()
        self.show_all()

    def _select_row_at_coords(self, x, y):
        """
        Select the row at the current cursor position.
        """
        wx, wy = self.tree.convert_bin_window_to_widget_coords(x, y)
        row = self.tree.get_dest_row_at_pos(wx, wy)
        if row:
            self.tree.get_selection().select_path(row[0])

    def _on_button_press(self, obj, event):
        """
        Handle button press, not double-click, that is done in init_interface
        """
        self._select_row_at_coords(event.x, event.y)
        if is_right_click(event):
            # ref = self.get_selected()
            # if ref:
            self.right_click(obj, event)
            return True
        elif event.type == Gdk.EventType.BUTTON_PRESS and event.button == 2:
            fun = self.get_middle_click()
            if fun:
                fun()
                return True
        return False

    def get_popup_menu_items(self):
        """
        Create the list needed to populate the right popup action
        An entry is
            ( needs_write_access, title, function)
        """
        if self.share_btn:
            itemlist = [
                (True, _("_Add"), self.add_button_clicked),
                (True, _("Share"), self.share_button_clicked),
                (False, _("_Edit"), self.edit_button_clicked),
                (True, _("_Remove"), self.del_button_clicked),
            ]
        else:
            itemlist = [
                (True, _("_Add"), self.add_button_clicked),
                (False, _("_Edit"), self.edit_button_clicked),
                (True, _("_Remove"), self.del_button_clicked),
            ]
        return itemlist

    def get_middle_click(self):
        return None

    def right_click(self, obj, event):
        """
        On right click show a popup menu.
        This is populated with get_popup_menu_items
        """
        self.__store_menu = Gtk.Menu()  # need to keep reference or menu disappears
        menu = self.__store_menu
        menu.set_reserve_toggle_size(False)
        for need_write, title, func in self.get_popup_menu_items():
            item = Gtk.MenuItem.new_with_mnemonic(title)
            item.connect("activate", func)
            if need_write and self.dbstate.db.readonly:
                item.set_sensitive(False)
            item.show()
            menu.append(item)
        menu.popup_at_pointer(event)
        return True

    def find_index(self, obj):
        """
        returns the index of the object within the associated data
        """
        return self.get_data().index(obj)

    def _set_dnd(self):
        """
        Set up drag-n-drop. The source and destination are set by calling .target()
        on the _DND_TYPE. Obviously, this means that there must be a _DND_TYPE
        variable defined that points to an entry in DdTargets.
        """

        if self._DND_EXTRA:
            dnd_types = [self._DND_TYPE.target(), self._DND_EXTRA.target()]
        else:
            dnd_types = [self._DND_TYPE.target()]

        self.tree.enable_model_drag_dest(dnd_types, Gdk.DragAction.COPY)
        self.tree.enable_model_drag_source(
            Gdk.ModifierType.BUTTON1_MASK,
            [self._DND_TYPE.target()],
            Gdk.DragAction.COPY,
        )

        self.tree.connect("drag_data_get", self.drag_data_get)
        self.tree.connect_after("drag-begin", self.after_drag_begin)
        if not self.dbstate.db.readonly:
            self.tree.connect("drag_data_received", self.drag_data_received)
            self.tree.connect("drag_motion", self.tree_drag_motion)

    def drag_data_get(self, widget, context, sel_data, info, time):
        """
        Provide the drag_data_get function, which passes a tuple consisting of:

           1) Drag type defined by the .drag_type field specified by the value
              assigned to _DND_TYPE
           2) The id value of this object, used for the purpose of determining
              the source of the object. If the source of the object is the same
              as the object, we are doing a reorder instead of a normal drag
              and drop
           3) Pickled data. The pickled version of the selected object
           4) Source row. Used for a reorder to determine the original position
              of the object
        """

        # get the selected object, returning if not is defined
        obj = self.get_selected()
        if not obj:
            return

        # pickle the data, and build the tuple to be passed
        value = (self._DND_TYPE.drag_type, id(self), obj, self.find_index(obj))
        data = pickle.dumps(value)

        # pass as a string (8 bits)
        sel_data.set(self._DND_TYPE.atom_drag_type, 8, data)

    def drag_data_received(self, widget, context, x, y, sel_data, info, time):
        """
        Handle the standard gtk interface for drag_data_received.

        If the selection data is defined, extract the value from sel_data.data,
        and decide if this is a move or a reorder.
        """
        if sel_data and sel_data.get_data():
            data = pickle.loads(sel_data.get_data())
            if isinstance(data, list):
                data = [pickle.loads(x) for x in data]
            else:
                data = [data]
            for value in data:
                (mytype, selfid, obj, row_from) = value

                # make sure this is the correct DND type for this object
                if mytype == self._DND_TYPE.drag_type:
                    # determine the destination row
                    row = self._find_row(x, y)

                    # if the is same object, we have a move, otherwise,
                    # it is a standard drag-n-drop

                    if id(self) == selfid and self.get_selected() is not None:
                        self._move(row_from, row, obj)
                    else:
                        self._handle_drag(row, obj)
                elif self._DND_EXTRA and mytype == self._DND_EXTRA.drag_type:
                    self.handle_extra_type(mytype, obj)
            self.rebuild()

    def tree_drag_motion(self, *args):
        """
        On drag motion one wants the list to show as the database
        representation so it is clear how save will change the data.
        """
        pass

    def after_drag_begin(self, widget, drag_context):
        """
        We want to show the icon during drag instead of the long row entry
        """
        Gtk.drag_set_icon_name(drag_context, self.get_icon_name(), 0, 0)

    def handle_extra_type(self, objtype, obj):
        pass

    def _find_row(self, x, y):
        row = self.tree.get_dest_row_at_pos(x, y)
        if row is None:
            return len(self.get_data())
        else:
            path = row[0].get_indices()
            if row[1] in (
                Gtk.TreeViewDropPosition.BEFORE,
                Gtk.TreeViewDropPosition.INTO_OR_BEFORE,
            ):
                return path[0]
            else:
                return path[0] + 1

    def _handle_drag(self, row, obj):
        self.get_data().insert(row, obj)
        self.changed = True

    def _move(self, row_from, row_to, obj):
        dlist = self.get_data()
        if row_from < row_to:
            dlist.insert(row_to, obj)
            del dlist[row_from]
        else:
            del dlist[row_from]
            dlist.insert(row_to, obj)
        self.changed = True

    def _move_up(self, row_from, obj, selmethod=None):
        """
        Move the item a position up in the EmbeddedList.
        Eg: 0,1,2,3 needs to become 0,2,1,3, here row_from = 2
        """
        if selmethod:
            dlist = selmethod()
        else:
            dlist = self.get_data()
        del dlist[row_from]
        dlist.insert(row_from - 1, obj)
        self.changed = True
        self.rebuild()
        # select the row
        path = "%d" % (row_from - 1)
        self.tree.get_selection().select_path(path)
        # The height/location of Gtk.treecells is calculated in an idle handler
        # so use idle_add to scroll cell into view.
        GLib.idle_add(self.tree.scroll_to_cell, path)

    def _move_down(self, row_from, obj, selmethod=None):
        """
        Move the item a position down in the EmbeddedList.
        Eg: 0,1,2,3 needs to become 0,2,1,3, here row_from = 1
        """
        if selmethod:
            dlist = selmethod()
        else:
            dlist = self.get_data()
        del dlist[row_from]
        dlist.insert(row_from + 1, obj)
        self.changed = True
        self.rebuild()
        # select the row
        path = "%d" % (row_from + 1)
        self.tree.get_selection().select_path(path)
        GLib.idle_add(self.tree.scroll_to_cell, path)

    def get_icon_name(self):
        """
        Specifies the basic icon used for a generic list. Typically,
        a derived class will override this. The icon chosen is the
        STOCK_JUSTIFY_FILL icon, which in the default GTK style
        looks kind of like a list.
        """
        return "format-justify-fill"

    def del_button_clicked(self, obj):
        ref = self.get_selected()
        if ref:
            ref_list = self.get_data()
            ref_list.remove(ref)
            self.changed = True
            self.rebuild()

    def up_button_clicked(self, obj):
        ref = self.get_selected()
        if ref:
            pos = self.find_index(ref)
            if pos > 0:
                self._move_up(pos, ref)

    def down_button_clicked(self, obj):
        ref = self.get_selected()
        if ref:
            pos = self.find_index(ref)
            if pos >= 0 and pos < len(self.get_data()) - 1:
                self._move_down(pos, ref)

    def build_interface(self):
        """
        Builds the interface, instantiating a Gtk.TreeView in a
        Gtk.ScrolledWindow.
        """

        # create the tree, turn on rule hinting and connect the
        # button press to the double click function.

        _name = self.get_config_name()
        self.tree = PersistentTreeView(self.uistate, _name)
        self.tree.set_reorderable(True)
        self.tree.connect("button_press_event", self.double_click)
        self.tree.connect("key_press_event", self.key_pressed)
        self.track_ref_for_deletion("tree")

        # create the scrolled window, and attach the treeview
        scroll = Gtk.ScrolledWindow()
        scroll.set_shadow_type(Gtk.ShadowType.IN)
        scroll.set_policy(Gtk.PolicyType.AUTOMATIC, Gtk.PolicyType.AUTOMATIC)
        scroll.add(self.tree)
        self.pack_end(scroll, True, True, 0)

    def get_selected(self):
        """
        Return the value associated with selected row in the model,
        based of the _HANDLE_COL value. Each model must define this
        to indicate what the returned value should be. If no selection
        has been made, None is returned.
        """
        (model, node) = self.selection.get_selected()
        if node:
            return model.get_value(node, self._HANDLE_COL)
        return None

    def is_empty(self):
        """
        Return True if the get_data returns a length greater than
        0. Typically, get_data returns the list of associated data.
        """
        return len(self.get_data()) == 0

    def get_data(self):
        """
        Return the data associated with the list. This is typically
        a list of objects.

        This should be overridden in the derived classes.
        """
        raise NotImplementedError

    def column_order(self):
        """
        Specifies the column order for the columns. This should be
        in the format of a list of tuples, in the format of (int,int),
        where the first in indicates if the column is visible, and the
        second column indicates the index into the model.

        This should be overridden in the derived classes.
        """
        raise NotImplementedError

    def setup_editable_col(self):
        """
        inherit this and set the variables needed for editable columns
        Variable edit_col_funcs needs to be a dictionary from model col_nr to
        function to call for
        Example:
        self.edit_col_funcs ={1: {'edit_start': self.on_edit_start,
                                  'edited': self.on_edited
                              }}
        """
        self.edit_col_funcs = {}

    def build_columns(self):
        """
        Builds the columns and inserts them into the TreeView. Any
        previous columns exist, they will be in the self.columns array,
        and removed.
        """

        # remove any existing columns, which would be stored in
        # self.columns

        list(map(self.tree.remove_column, self.columns))
        self.columns = []
        self.setup_editable_col()

        # loop through the values returned by column_order
        for pair in self.column_order():
            # if the first value isn't 1, then we skip the values
            if not pair[0]:
                continue

            # extract the name from the _column_names variable, and
            # assign it to the column name. The text value is extracted
            # from the model column specified in pair[1]
            name = self._column_names[pair[1]][0]
            col_icon = self._column_names[pair[1]][5]
            model_col = self._column_names[pair[1]][1]
            type_col = self._column_names[pair[1]][3]

            if type_col in [TEXT_COL, MARKUP_COL, TEXT_EDIT_COL]:
                if type_col == TEXT_EDIT_COL:
                    renderer = CellRendererTextEdit()
                else:
                    renderer = Gtk.CellRendererText()
                renderer.set_property("ellipsize", Pango.EllipsizeMode.END)
                if type_col == TEXT_COL or type_col == TEXT_EDIT_COL:
                    column = Gtk.TreeViewColumn(name, renderer, text=pair[1])
                else:
                    column = Gtk.TreeViewColumn(name, renderer, markup=pair[1])
                if not self._column_names[pair[1]][4] == -1:
                    # apply weight attribute
                    column.add_attribute(
                        renderer, "weight", self._column_names[pair[1]][4]
                    )
                # set up editable
                if type_col == TEXT_EDIT_COL:
                    # model col must have functions defined
                    callbacks = self.edit_col_funcs[model_col]
                    for renderer in column.get_cells():
                        renderer.set_property("editable", not self.dbstate.db.readonly)
                        renderer.connect(
                            "editing_started", callbacks["edit_start"], model_col
                        )
                        renderer.connect("edited", callbacks["edited"], model_col)
            elif self._column_names[pair[1]][3] == ICON_COL:
                self.col_icons[pair[1]] = col_icon
                self.pb_renderer = Gtk.CellRendererPixbuf()
                column = Gtk.TreeViewColumn(name, self.pb_renderer)
                column.set_cell_data_func(self.pb_renderer, self.icon_func, pair[1])
            else:
                raise NotImplementedError(
                    "Unknown column type: %s, with column name %s"
                    % (type_col, self._column_names[pair[1]][3])
                )
            if col_icon is not None:
                image = Gtk.Image()
                image.set_from_icon_name(col_icon, Gtk.IconSize.MENU)
                image.set_tooltip_text(name)
                image.show()
                column.set_widget(image)
                column.set_resizable(False)
            else:
                # insert the colum into the tree
                column.set_resizable(True)
            column.set_clickable(True)
            if self._column_names[pair[1]][2] != -1:
                column.set_sizing(Gtk.TreeViewColumnSizing.FIXED)
                # column.set_min_width(self._column_names[pair[1]][2])
                column.set_fixed_width(self._column_names[pair[1]][2])
            else:
                column.set_expand(True)

            column.set_sort_column_id(self._column_names[pair[1]][1])
            self.columns.append(column)
            self.tree.append_column(column)
        self.track_ref_for_deletion("columns")
        self.tree.restore_column_size()

    def get_config_name(self):
        """used to associate the selector config name to the PersistentTreeView"""
        return self.config_key

    def icon_func(self, column, renderer, model, iter_, col_num):
        """
        Set the stock icon property of the cell renderer.  We use a cell data
        function because there is a problem returning None from a model.
        """
        icon_name = model.get_value(iter_, col_num)
        if icon_name == "" or icon_name == False:
            icon_name = None
        elif icon_name == True:
            icon_name = self.col_icons[col_num]
        renderer.set_property("icon-name", icon_name)

    def construct_model(self):
        """
        Method that creates the model using the passed build_model parameter
        """
        return self.build_model(self.get_data(), self.dbstate.db)

    def rebuild(self):
        """
        Rebuilds the data in the database by creating a new model,
        using the build_model function passed at creation time.
        """
        offset = self.tree.get_visible_rect()
        # during rebuild, don't do _selection_changed
        self.dirty_selection = True
        (model, node) = self.selection.get_selected()
        selectedpath = None
        if node:
            selectedpath = model.get_path(node)
        if self.model and hasattr(self.model, "destroy"):
            self.tree.set_model(None)
            self.model.destroy()
        try:
            self.model = self.construct_model()
        except AttributeError as msg:
            from ...dialog import RunDatabaseRepair
            import traceback

            traceback.print_exc()
            RunDatabaseRepair(str(msg), parent=self.uistate.window)
            return

        self.tree.set_model(self.model)
        # reset previous select
        if selectedpath is not None:
            self.selection.select_path(selectedpath)
        # self.selection.select_path(node)
        self._set_label()
        # model and tree are reset, allow _selection_changed again, and force it
        self.dirty_selection = False
        self._selection_changed()
        if self.tree.get_realized():
            GLib.idle_add(self.tree.scroll_to_point, offset.x, offset.y)
        self.post_rebuild(selectedpath)

    def post_rebuild(self, prebuildpath):
        """
        Allow post rebuild embeddedlist specific handling.
        @param prebuildpath: path selected before rebuild, None if none
        @type prebuildpath: tree path
        """
        pass

    def rebuild_callback(self):
        """
        The view must be remade when data changes outside this tab.
        Use this method to connect to after a db change. It makes sure the
        data is obtained again from the present object and the db what is not
        present in the obj, and the view rebuild
        """
        self.changed = True
        self.rebuild()



#
# Gramps - a GTK+/GNOME based genealogy program
#
# Copyright (C) 2010  Doug Blank <doug.blank@gmail.com>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
#

"""
Display link references for a note
"""

from gramps.gen.simple import SimpleAccess, SimpleDoc
from gramps.gui.plug.quick import QuickTable
from gramps.gen.lib import StyledTextTagType
from gramps.gen.const import GRAMPS_LOCALE as glocale

_ = glocale.translation.gettext


def run(database, document, obj):
    """
    Display link references for this note.
    """

    # setup the simple access functions
    sdb = SimpleAccess(database)
    sdoc = SimpleDoc(document)
    stab = QuickTable(sdb)

    # display the title
    sdoc.title(_("Link References for this note"))
    sdoc.paragraph("\n")
    stab.columns(_("Type"), _("Reference"), _("Link check"))

    for ldomain, ltype, lprop, lvalue in obj.get_links():
        if ldomain == "gramps":
            tagtype = _(ltype)
            ref_obj = sdb.get_link(ltype, lprop, lvalue)
            if ref_obj:
                tagvalue = ref_obj
                tagcheck = _("Ok")
            else:
                tagvalue = lvalue
                tagcheck = _("Failed: missing object")
        else:
            tagtype = _("Internet")
            tagvalue = lvalue
            tagcheck = ""
        stab.row(tagtype, tagvalue, tagcheck)

    if stab.get_row_count() > 0:
        stab.write(sdoc)
        document.has_data = True
    else:
        sdoc.paragraph(_("No link references for this note"))
        sdoc.paragraph("")
        document.has_data = False
    sdoc.paragraph("")



# -*- coding: utf-8 -*-
#
# Gramps - a GTK+/GNOME based genealogy program
#
# Copyright (C) 2003-2005  Donald N. Allingham
# Copyright (C) 2008       Brian G. Matherly
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#

# Rewritten in 2008 for 3.x version by Łukasz Rymarczyk
# Written in 2007 by Piotr Czubaszek, largely based on rel_de.py by Alex Roitman.

# PL: Po objaśnienia oznaczania relacji zobacz relationship.py
# EN: For more information see relationship.py
#
"""
Polish-specific definitions of relationships.
"""

# -------------------------------------------------------------------------
#
# Gramps modules
#
# -------------------------------------------------------------------------

from gramps.gen.lib import Person
import gramps.gen.relationship

# -------------------------------------------------------------------------


# określa liczebnik porządkowy

_level_name = [
    "pierwszego",
    "drugiego",
    "trzeciego",
    "czwartego",
    "piątego",
    "szóstego",
    "siódmego",
    "ósmego",
    "dziewiątego",
    "dziesiątego",
    "jedenastego",
    "dwunastego",
    "trzynastego",
    "czternastego",
    "piętnastego",
    "szesnastego",
    "siedemnastego",
    "osiemnastego",
    "dziewiętnastego",
    "dwudziestego",
]


_father_level = [
    "",
    "ojciec",
    "dziadek",
    "pradziadek",
    "prapradziadek",
    "praprapradziadek",
    "prapraprapradziadek",
    "praprapraprapradziadek",
    "prapraprapraprapradziadek",
    "praprapraprapraprapradziadek",
    "prapraprapraprapraprapradziadek",
]


_mother_level = [
    "",
    "matka",
    "babcia",
    "prababcia",
    "praprababcia",
    "prapraprababcia",
    "praprapraprababcia",
    "prapraprapraprababcia",
    "praprapraprapraprababcia",
    "prapraprapraprapraprababcia",
    "praprapraprapraprapraprababcia",
]

_son_level = [
    "",
    "syn",
    "wnuk",
    "prawnuk",
    "praprawnuk",
    "prapraprauwnuk",
    "praprapraprauwnuk",
    "prapraprapraprawnuk",
    "praprapraprapraprawnuk",
    "prapraprapraprapraprawnuk",
    "praprapraprapraprapraprawnuk",
]

_daughter_level = [
    "",
    "córka",
    "wnuczka",
    "prawnuczka",
    "praprawnuczka",
    "prapraprauwnuczka",
    "praprapraprauwnuczka",
    "prapraprapraprawnuczka",
    "praprapraprapraprawnuczka",
    "prapraprapraprapraprawnuczka",
    "praprapraprapraprapraprawnuczka",
]

_sister_level_of_male = [
    "",
    "siostra",
    "ciotka stryjeczna",
    "babcia stryjeczna",
    "prababcia stryjeczna",
    "praprababcia stryjeczna",
    "prapraprababcia stryjeczna",
    "praprapraprababcia stryjeczna",
    "prapraprapraprababcia stryjeczna",
    "praprapraprapraprababcia stryjeczna",
    "prapraprapraprapraprababcia stryjeczna",
    "praprapraprapraprapraprababcia stryjeczna",
]

_sister_level_of_female = [
    "",
    "siostra",
    "ciotka",
    "babcia cioteczna",
    "prababcia cioteczna",
    "praprababcia cioteczna",
    "prapraprababcia cioteczna",
    "praprapraprababcia cioteczna",
    "prapraprapraprababcia cioteczna",
    "praprapraprapraprababcia cioteczna",
    "prapraprapraprapraprababcia cioteczna",
    "praprapraprapraprapraprababcia cioteczna",
]

_brother_level_of_male = [
    "",
    "brat",
    "stryj",
    "dziadek stryjeczny",
    "pradziadek stryjeczny",
    "prapradziadek stryjeczny",
    "praprapradziadek stryjeczny",
    "prapraprapradziadek stryjeczny",
    "praprapraprapradziadek stryjeczny",
    "prapraprapraprapradziadek stryjeczny",
    "praprapraprapraprapradziadek stryjeczny",
    "prapraprapraprapraprapradziadek stryjeczny",
]

_brother_level_of_female = [
    "",
    "brat",
    "wuj",
    "dziadek cioteczny",
    "pradziadek cioteczny",
    "prapradziadek cioteczny",
    "praprapradziadek cioteczny",
    "prapraprapradziadek cioteczny",
    "praprapraprapradziadek cioteczny",
    "prapraprapraprapradziadek cioteczny",
    "praprapraprapraprapradziadek cioteczny",
    "prapraprapraprapraprapradziadek cioteczny",
]

_nephew_level_of_brothers_son = [
    "",
    "bratanek",
    "syn bratanka",
    "wnuk bratanka",
    "prawnuk bratanka",
    "praprawnuk bratanka",
    "prapraprawnuk bratanka",
    "praprapraprawnuk bratanka",
    "prapraprapraprawnuk bratanka",
    "praprapraprapraprawnuk bratanka",
    "prapraprapraprapraprawnuk bratanka",
]

_nephew_level_of_brothers_daughter = [
    "",
    "bratanica",
    "syn bratanicy",
    "wnuk bratanicy",
    "prawnuk bratanicy",
    "praprawnuk bratanicy",
    "prapraprawnuk bratanicy",
    "praprapraprawnuk bratanicy",
    "prapraprapraprawnuk bratanicy",
    "praprapraprapraprawnuk bratanicy",
    "prapraprapraprapraprawnuk bratanicy",
    "praprapraprapraprapraprawnuk bratanicy",
]

_nephew_level_of_sisters_son = [
    "",
    "siostrzeniec",
    "syn siostrzeńca",
    "wnuk siostrzeńca",
    "prawnuk siostrzeńca",
    "praprawnuk siostrzeńca",
    "prapraprawnuk siostrzeńca",
    "praprapraprawnuk siostrzeńca",
    "prapraprapraprawnuk siostrzeńca",
    "praprapraprapraprawnuk siostrzeńca",
    "prapraprapraprapraprawnuk siostrzeńca",
]

_nephew_level_of_sisters_daughter = [
    "",
    "siostrzenica",
    "syn siostrzenicy",
    "wnuk siostrzenicy",
    "prawnuk siostrzenicy",
    "praprawnuk siostrzenicy",
    "prapraprawnuk siostrzenicy",
    "praprapraprawnuk siostrzenicy",
    "prapraprapraprawnuk siostrzenicy",
    "praprapraprapraprawnuk siostrzenicy",
    "prapraprapraprapraprawnuk siostrzenicy",
]

_niece_level_of_brothers_son = [
    "",
    "bratanica",
    "córka bratanka",
    "wnuczka bratanka",
    "prawnuczka bratanka",
    "praprawnuczka bratanka",
    "prapraprawnuczka bratanka",
    "praprapraprawnuczka bratanka",
    "prapraprapraprawnuczka bratanka",
    "praprapraprapraprawnuczka bratanka",
]

_niece_level_of_brothers_daughter = [
    "",
    "bratanica",
    "córka bratanicy",
    "wnuczka bratanicy",
    "prawnuczka bratanicy",
    "praprawnuczka bratanicy",
    "prapraprawnuczka bratanicy",
    "praprapraprawnuczka bratanicy",
    "prapraprapraprawnuczka bratanicy",
    "praprapraprapraprawnuczka bratanicy",
]

_niece_level_of_sisters_son = [
    "",
    "siostrzenica",
    "córka siostrzeńca",
    "wnuczka siostrzeńca",
    "prawnuczka siostrzeńca",
    "praprawnuczka siostrzeńca",
    "prapraprawnuczka siostrzeńca",
    "praprapraprawnuczka siostrzeńca",
    "prapraprapraprawnuczka siostrzeńca",
    "praprapraprapraprawnuczka siostrzeńca",
]

_niece_level_of_sisters_daughter = [
    "",
    "siostrzenica",
    "córka siostrzenicy",
    "wnuczka siostrzenicy",
    "prawnuczka siostrzenicy",
    "praprawnuczka siostrzenicy",
    "prapraprawnuczka siostrzenicy",
    "praprapraprawnuczka siostrzenicy",
    "prapraprapraprawnuczka siostrzenicy",
    "praprapraprapraprawnuczka siostrzenicy",
]

# -------------------------------------------------------------------------
#
#
#
# -------------------------------------------------------------------------


class RelationshipCalculator(gramps.gen.relationship.RelationshipCalculator):
    """
    RelationshipCalculator Class
    """

    def __init__(self):
        gramps.gen.relationship.RelationshipCalculator.__init__(self)

    def get_son(self, level, inlaw=""):
        """
        Podaje tekst zawierający informację, jak bardzo potomek męski
        (np. syn) jest spokrewniony do danej osoby
        """

        # Określ, czy osoba jest przybraną, czy rodzoną
        if inlaw == "":
            t_inlaw = ""
        else:
            t_inlaw = "przybrany "

        # TODO: dodać rozpoznawanie pasierb/pasierbica
        if level >= 0 and level < len(_son_level):
            return t_inlaw + _son_level[level]

        elif level >= len(_son_level) and (level - 1) < len(_level_name):
            return t_inlaw + "potomek męski %s pokolenia" % _level_name[level - 1]

        else:
            return t_inlaw + "potomek męski w %d pokoleniu" % level

    def get_daughter(self, level, inlaw=""):
        """
        Podaje tekst zawierający informację, jak bardzo potomek żeński
        (np. córka) jest spokrewniony do danej osoby
        """

        # Określ, czy osoba jest przybraną, czy rodzoną
        #   + stwórz obie formy (męską i żeńską)
        if inlaw == "":
            t_inlaw = ""
            t_inlawM = ""
        else:
            t_inlaw = "przybrana "
            t_inlawM = "przybrany "

        # TODO: dodać rozpoznawanie pasierb/pasierbica
        if level >= 0 and level < len(_daughter_level):
            return t_inlaw + _daughter_level[level]

        elif level >= len(_daughter_level) and (level - 1) < len(_level_name):
            return t_inlawM + "potomek żeński %s pokolenia" % _level_name[level - 1]
        else:
            return t_inlawM + "potomek żeński w %d pokoleniu" % level

    def get_child_unknown(self, level, inlaw=""):
        """
        Podaje tekst zawierający informację, jak bardzo potomek
        o nieokreślonej płci jest spokrewniony dodanej osoby
        """

        # Określ, czy osoba jest przybraną, czy rodzoną
        if inlaw == "":
            t_inlaw = ""
        else:
            t_inlaw = "przybrany "

        if level == 1:
            if inlaw == "":
                return "dziecko"
            else:
                return "przybrane dziecko"

        elif level >= 1 and (level - 1) < len(_level_name):
            return t_inlaw + "potomek %s pokolenia" % _level_name[level - 1]

        else:
            return t_inlaw + "potomek w %d pokoleniu" % level

    def get_sword_distaff(self, level, reltocommon, spacebefore=""):
        """
        PL: Generuje relację po mieczu/po kądzieli
        EN: Generate relation 'by sword' or 'by distaff', polish specific
        """
        if level <= 1:
            return ""

        elif level == 2:
            # dziadek/babcia

            if reltocommon[0] == self.REL_FATHER:
                # ze strony rodzonego ojca
                return spacebefore + "po mieczu"
            elif reltocommon[0] == self.REL_MOTHER:
                # ze strony rodzonej matki
                return spacebefore + "po kądzieli"
            else:
                # relacja inna niż rodzona
                return ""

        elif level == 3:
            # pradziadek/prababcia

            if (reltocommon[0] == self.REL_FATHER) & (
                reltocommon[1] == self.REL_FATHER
            ):
                # pradziadek od dziadka ze strony ojca
                return spacebefore + "podwójnego miecza"

            elif (reltocommon[0] == self.REL_FATHER) & (
                reltocommon[1] == self.REL_MOTHER
            ):
                # pradziadek od babci ze strony ojca
                return spacebefore + "raz po mieczu, dalej po kądzieli"

            elif (reltocommon[0] == self.REL_MOTHER) & (
                reltocommon[1] == self.REL_FATHER
            ):
                # pradziadek od dziadka ze strony matki
                return spacebefore + "raz po kądzieli, dalej po mieczu"

            elif (reltocommon[0] == self.REL_MOTHER) & (
                reltocommon[1] == self.REL_MOTHER
            ):
                # pradziadek od babci ze strony matki
                return spacebefore + "podwójnej kądzieli"

            else:
                # relacja inna niż rodzona
                return ""

        elif level == 4:
            # prapradziadek/praprababcia

            if (
                (reltocommon[0] == self.REL_FATHER)
                & (reltocommon[1] == self.REL_FATHER)
                & (reltocommon[2] == self.REL_FATHER)
            ):
                # tzw. linia męska
                return spacebefore + "potrójnego miecza"

            if (
                (reltocommon[0] == self.REL_FATHER)
                & (reltocommon[1] == self.REL_FATHER)
                & (reltocommon[2] == self.REL_FATHER)
            ):
                # tzw. linia żeńska
                return spacebefore + "potrójnego miecza"

            else:
                return ""

        else:
            return ""

    def get_father(self, level, reltocommon, inlaw=""):
        """
        Podaje tekst zawierający informację, jak bardzo przodek męski
        (np. ojciec) jest spokrewniony do danej osoby
        """
        if inlaw == "":
            t_inlaw = ""
        else:
            t_inlaw = "przybrany "

        if level >= 0 and level < len(_father_level):
            # Jeśli znasz bezpośrednią nazwę relacji, to ją zastosuj

            if level == 1:
                # ojciec
                return t_inlaw + _father_level[level]

            elif (level >= 2) & (level <= 4):
                # dziadek, pradziadek, prapradziadek
                return (
                    t_inlaw
                    + _father_level[level]
                    + self.get_sword_distaff(level, reltocommon, " ")
                )

            else:
                return t_inlaw + _father_level[level]

        elif level >= len(_father_level) and (level - 1) < len(_level_name):
            # jeśli istnieje liczebnik dla danej liczby
            return t_inlaw + "przodek męski %s pokolenia" % (_level_name[level - 1])

        else:
            # dla pozostałych przypadków wypisz relację liczbowo
            return t_inlaw + "przodek męski w %d pokoleniu" % level

    def get_mother(self, level, reltocommon, inlaw=""):
        """
        Podaje tekst zawierający informację, jak bardzo przodek żeński
        (np. matka) jest spokrewniony do danej osoby
        """

        if inlaw == "":
            t_inlaw = ""
        else:
            t_inlaw = "przybrana "

        if level >= 0 and level < len(_mother_level):
            # Jeśli znasz bezpośrednią nazwę relacji, to ją zastosuj

            if level == 1:
                # matka
                return t_inlaw + _mother_level[level]

            elif (level >= 2) & (level <= 4):
                # babcia, prababcia, praprababcia
                return (
                    t_inlaw
                    + _mother_level[level]
                    + self.get_sword_distaff(level, reltocommon, " ")
                )

            else:
                return t_inlaw + _mother_level[level]

        elif level >= len(_mother_level) and (level - 1) < len(_level_name):
            # jeśli istnieje liczebnik dla danej liczby
            return t_inlaw + "przodek żeński %s pokolenia" % (_level_name[level - 1])

        else:
            # dla pozostałych przypadków wypisz relację liczbowo
            return t_inlaw + "przodek żeński w %d pokoleniu" % level

    def get_parent_unknown(self, level, inlaw=""):
        """
        Podaje tekst zawierający informację, jak bardzo przodek
        o nieokreślonej płci jest spokrewniony dodanej osoby
        """

        if inlaw == "":
            t_inlaw = ""
        else:
            t_inlaw = "przybrany "

        if level == 1:
            return t_inlaw + "rodzic"

        elif level > 1 and (level - 1) < len(_level_name):
            if (level >= 2) & (level <= 4):
                # babcia, prababcia, praprababcia
                # (albo dziadek, pradziadek, prapradziadek)
                tmp = t_inlaw + "przodek %s pokolenia" % (_level_name[level - 1])
                # TODO: try to recognize a gender...
                return tmp
                # +  self.get_sword_distaff(level, reltocommon, ' ')
            else:
                return t_inlaw + "przodek %s pokolenia" % (_level_name[level - 1])
        else:
            return t_inlaw + "przodek w %d pokoleniu" % level

    def get_single_relationship_string(
        self,
        Ga,
        Gb,
        gender_a,
        gender_b,
        reltocommon_a,
        reltocommon_b,
        only_birth=True,
        in_law_a=False,
        in_law_b=False,
    ):
        """
        Provide a string that describes the relationsip between a person, and
        another person. E.g. "grandparent" or "child".
        """

        if only_birth:
            step = ""
        else:
            step = self.STEP

        if in_law_a or in_law_b:
            inlaw = self.INLAW
        else:
            inlaw = ""

        # b is the same person as a
        if Ga == Gb == 0:
            rel_str = "ta sama osoba"

        elif Ga == 0:
            # b is son/descendant of a

            if gender_b == Person.MALE:
                if inlaw and Gb == 1 and not step:
                    rel_str = "zięć"
                else:
                    rel_str = self.get_son(Gb, inlaw)

            elif gender_b == Person.FEMALE:
                if inlaw and Gb == 1 and not step:
                    rel_str = "synowa"
                else:
                    rel_str = self.get_daughter(Gb, inlaw)

            else:
                rel_str = self.get_child_unknown(Gb, inlaw)

        elif Gb == 0:
            # b is parent/grand parent of a

            if gender_b == Person.MALE:
                if inlaw and Gb == 1 and not step:
                    # TODO: znaleźć odpowiedniki w zależności czy to syn/córka
                    rel_str = "teść"
                else:
                    rel_str = self.get_father(Ga, reltocommon_a, inlaw)

            elif gender_b == Person.FEMALE:
                if inlaw and Gb == 1 and not step:
                    # TODO: znaleźć odpowiedniki w zależności czy to syn/córka
                    rel_str = "teściowa"
                else:
                    rel_str = self.get_mother(Ga, reltocommon_a, inlaw)

            else:
                rel_str = self.get_parent_unknown(Ga, inlaw)

        elif Ga == Gb == 1:
            # rodzeństwo
            if gender_b == Person.MALE:
                if inlaw and not step:
                    rel_str = "brat przyrodni"
                else:
                    rel_str = "brat rodzony"

            elif gender_b == Person.FEMALE:
                if inlaw and not step:
                    rel_str = "siostra przyrodnia"
                else:
                    rel_str = "siostra rodzony"
            else:
                rel_str = "brat/siostra"

        elif Gb == 1 and Ga > 1:
            # Przyjmij, że nie rozróżniamy osób prawnie i nieprawnie przybranych...

            if Ga == 2:
                # rodzeństwo rodziców

                # brat ojca, czyli stryj
                if (gender_b == Person.MALE) & (reltocommon_a[0] == self.REL_FATHER):
                    rel_str = "stryj"

                # siostra ojca, czyli ciotka ???
                elif (gender_b == Person.FEMALE) & (
                    reltocommon_a[0] == self.REL_FATHER
                ):
                    rel_str = "ciotka (tzw. stryjna)"

                # brat matki, czyli wuj/wujek
                elif (gender_b == Person.MALE) & (reltocommon_a[0] == self.REL_MOTHER):
                    rel_str = "wuj (wujek)"

                # siostra matki, czyli ciotka
                elif (gender_b == Person.FEMALE) & (
                    reltocommon_a[0] == self.REL_MOTHER
                ):
                    rel_str = "ciotka"

                else:
                    rel_str = "brat lub siostra rodzica"

            elif Ga == 3:
                # rodzeństwo dziadków rodziców osoby sprawdzanej

                # rodzeństwo dziadka po mieczu (ojca ojca)
                if (reltocommon_a[0] == self.REL_FATHER) & (
                    reltocommon_a[1] == self.REL_FATHER
                ):
                    if gender_b == Person.MALE:
                        rel_str = "dziadek stryjeczny (tzw przestryj, stary stryj)"
                    elif gender_b == Person.FEMALE:
                        rel_str = "babcia stryjeczna"
                    else:
                        rel_str = "rodzeństwo przodka w 2 pokoleniu"

                # rodzeństwo babki po mieczu (matki ojca)
                elif (reltocommon_a[0] == self.REL_FATHER) & (
                    reltocommon_a[1] == self.REL_MOTHER
                ):
                    # TODO: Należy sprawdzić, czy w staropolszczyźnie nie ma
                    #            dokładniejszych określeń dla tego typu relacji
                    # TODO: EN: Try to check, whether in old polish language
                    #      are more specific word for this kind of relation
                    if gender_b == Person.MALE:
                        rel_str = "dziadek stryjeczny (tzw przestryj, stary stryj)"
                    elif gender_b == Person.FEMALE:
                        rel_str = "babcia stryjeczna"
                    else:
                        rel_str = "rodzeństwo przodka w 2 pokoleniu"

                # rodzeństwo dziadka po kądzieli (ojca matki)
                elif (reltocommon_a[0] == self.REL_MOTHER) & (
                    reltocommon_a[1] == self.REL_FATHER
                ):
                    # TODO: Należy sprawdzić, czy w staropolszczyźnie nie ma
                    #            dokładniejszych określeń dla tego typu relacji
                    # TODO: EN: Try to check, whether in old polish language
                    #      are more specific word for this kind of relation
                    if gender_b == Person.MALE:
                        rel_str = "dziadek cioteczny (starop. prapociot)"
                    elif gender_b == Person.FEMALE:
                        rel_str = "babcia cioteczna (starop. praciota)"
                    else:
                        rel_str = "rodzeństwo przodka w 2 pokoleniu"

                # rodzeństwo babki po kądzieli (matki matki)
                elif (reltocommon_a[0] == self.REL_MOTHER) & (
                    reltocommon_a[1] == self.REL_MOTHER
                ):
                    # TODO: Należy sprawdzić, czy w staropolszczyźnie nie ma
                    #           dokładniejszych określeń dla tego typu relacji
                    # TODO: EN: Try to check, whether in old polish language
                    #      are more specific word for this kind of relation
                    if gender_b == Person.MALE:
                        rel_str = "dziadek cioteczny (starop. prapociot)"
                    elif gender_b == Person.FEMALE:
                        rel_str = "babcia cioteczna  (starop. praciota)"
                    else:
                        rel_str = "rodzeństwo przodka w 2 pokoleniu"

                else:
                    if gender_b == Person.MALE:
                        rel_str = "rodzeństwo dziadka"
                    elif gender_b == Person.FEMALE:
                        rel_str = "rodzeństwo babci"
                    else:
                        rel_str = "rodzeństwo przodka w 2 pokoleniu"

            elif Ga > 3:
                # pradziadkowie...  (grandparents)

                if (gender_b == Person.MALE) & (reltocommon_a[0] == self.REL_FATHER):
                    if Ga >= 0 and Ga < len(_brother_level_of_male):
                        rel_str = _brother_level_of_male[Ga]
                    else:
                        rel_str = "rodzeństwo przodka męskiego %d pokolenia" % Ga

                elif (gender_b == Person.FEMALE) & (
                    reltocommon_a[0] == self.REL_FATHER
                ):
                    if Ga >= 0 and Ga < len(_sister_level_of_male):
                        rel_str = _sister_level_of_male[Ga]
                    else:
                        rel_str = "rodzeństwo przodka żeńskiego %d pokolenia" % Ga

                elif (gender_b == Person.MALE) & (reltocommon_a[0] == self.REL_MOTHER):
                    if Ga >= 0 and Ga < len(_brother_level_of_female):
                        rel_str = _brother_level_of_male[Ga]
                    else:
                        rel_str = "rodzeństwo przodka męskiego %d pokolenia" % Ga

                elif (gender_b == Person.FEMALE) & (
                    reltocommon_a[0] == self.REL_MOTHER
                ):
                    if Ga >= 0 and Ga < len(_sister_level_of_female):
                        rel_str = _sister_level_of_male[Ga]
                    else:
                        rel_str = "rodzeństwo przodka żeńskiego %d pokolenia" % Ga

                else:
                    rel_str = "rodzeństwo przodka %d pokolenia" % Ga
            else:
                # A program should never goes there, but...
                rel_str = "Relacja nie określona"

        elif Ga == 1 and Gb > 1:
            # syn brata
            if (gender_b == Person.MALE) & (reltocommon_b[0] == self.REL_FATHER):
                if Gb < len(_nephew_level_of_brothers_son):
                    rel_str = _nephew_level_of_brothers_son[Gb]
                else:
                    rel_str = "męski potomek w %d pokoleniu brata" % Gb

            # córka brata
            elif (gender_b == Person.FEMALE) & (reltocommon_b[0] == self.REL_FATHER):
                if Gb < len(_nephew_level_of_brothers_daughter):
                    rel_str = _nephew_level_of_brothers_daughter[Gb]
                else:
                    rel_str = "żeński potomek w %d pokoleniu brata" % Gb

            # syn siostry
            if (gender_b == Person.MALE) & (reltocommon_b[0] == self.REL_MOTHER):
                if Gb < len(_nephew_level_of_sisters_son):
                    rel_str = _nephew_level_of_sisters_son[Gb]
                else:
                    rel_str = "męski potomek w %d pokoleniu brata" % Gb

            # córka siostry
            elif (gender_b == Person.FEMALE) & (reltocommon_b[0] == self.REL_MOTHER):
                if Gb < len(_nephew_level_of_sisters_daughter):
                    rel_str = _nephew_level_of_sisters_daughter[Gb]
                else:
                    rel_str = "żeński potomek w %d pokoleniu brata" % Gb

            # potomek brata
            elif reltocommon_b[0] == self.REL_FATHER:
                rel_str = "potomek w %d pokoleniu brata" % Gb

            # potomek brata
            elif reltocommon_b[0] == self.REL_MOTHER:
                rel_str = "potomek w %d pokoleniu brata" % Gb

            else:
                rel_str = "potomek w %d pokoleniu rodzeństwa" % Gb

        elif Ga > 1 and Gb > 1:
            if gender_b == Person.MALE:
                if Ga == 2 and Gb == 2:
                    rel_str = "kuzyn"
                else:
                    rel_str = "daleki kuzyn (%d. stopień pokrewieństwa)" % (Ga + Gb)

            elif gender_b == Person.FEMALE:
                if Ga == 2 and Gb == 2:
                    rel_str = "kuzynka"
                else:
                    rel_str = "daleka kuzynka (%d. stopień pokrewieństwa)" % (Ga + Gb)

            else:
                if Ga == 2 and Gb == 2:
                    rel_str = "kuzyn(ka)"
                else:
                    rel_str = "daleki członek rodziny (%d. stopień pokrewieństwa)" % (
                        Ga + Gb
                    )

        else:
            # A program should never goes there, but...
            rel_str = "nieokreślony stopień pokrewieństwa"

        return rel_str

    def get_plural_relationship_string(
        self,
        Ga,
        Gb,
        reltocommon_a="",
        reltocommon_b="",
        only_birth=True,
        in_law_a=False,
        in_law_b=False,
    ):
        """
        Generate a text with information, how far away is a group of persons
        from a main person
        """

        if Ga == Gb == 0:
            return "ta sama osoba"
        if 0 == Ga:
            if 1 == Gb:
                return "Dzieci"
            if 2 == Gb:
                return "Wnuki"
            if 3 == Gb:
                return "Prawnuki"
            if 4 == Gb:
                return "Praprawnuki"
            return "Praprapra(n)wnuki"
        if 0 == Gb:
            if 1 == Ga:
                return "Rodzice"
            if 2 == Ga:
                return "Dziadkowie"
            if 3 == Ga:
                return "Pradziadkowie"
            if 4 == Ga:
                return "Praprapradziadkowie"
            return "Praprapra(n)dziadkowie"
        if 1 == Ga == Gb:
            return "Rodzeństwo"
        if 1 == Gb and Ga > 1:
            return "Wujowie/stryjowie i ciocie"
        if 1 < Gb and 1 == Ga:
            return "bratankowie(ice)/siostrzeńcy(nice)"
        if 1 < Ga and 1 < Gb:
            return "dalsza rodzina"
        return "relacja nieznana"

    def get_sibling_relationship_string(
        self, sib_type, gender_a, gender_b, in_law_a=False, in_law_b=False
    ):
        if in_law_a or in_law_b:
            inlaw = self.INLAW
        else:
            inlaw = ""

        if sib_type == self.NORM_SIB:
            if not inlaw:
                if gender_b == Person.MALE:
                    rel_str = "brat (rodzony)"
                elif gender_b == Person.FEMALE:
                    rel_str = "siostra (rodzona)"
                else:
                    rel_str = "brat lub siostra (rodzeni)"
            else:
                if gender_b == Person.MALE:
                    # TODO: znaleźć odpowiednik
                    rel_str = "brat (pasierb)"
                elif gender_b == Person.FEMALE:
                    # TODO: znaleźć odpowiednik
                    rel_str = "siostra (pasierbica)"
                else:
                    # TODO: znaleźć odpowiednik
                    rel_str = "brat lub siostra (pasierb/pasierbica)"
        elif sib_type == self.UNKNOWN_SIB:
            if not inlaw:
                if gender_b == Person.MALE:
                    rel_str = "brat"
                elif gender_b == Person.FEMALE:
                    rel_str = "siostra"
                else:
                    rel_str = "brat lub siostra"
            else:
                if gender_b == Person.MALE:
                    # TODO: znaleźć odpowiednik
                    rel_str = "brat (brat/szwagier)"
                elif gender_b == Person.FEMALE:
                    # TODO: znaleźć odpowiednik
                    rel_str = "siostra (bratowa/szwagierka)"
                else:
                    # TODO: znaleźć odpowiednik
                    rel_str = "brat lub siostra (szwagier/szagierka)"
        elif sib_type == self.HALF_SIB_FATHER:
            if gender_b == Person.MALE:
                rel_str = "brat przyrodni"
            elif gender_b == Person.FEMALE:
                rel_str = "siostra przyrodnia"
            else:
                rel_str = "brat/siostra przyrodni"
        elif sib_type == self.HALF_SIB_MOTHER:
            if gender_b == Person.MALE:
                rel_str = "brat przyrodni"
            elif gender_b == Person.FEMALE:
                rel_str = "siostra przyrodnia"
            else:
                rel_str = "brat/siostra przyrodni"
        elif sib_type == self.STEP_SIB:
            if gender_b == Person.MALE:
                rel_str = "brat przyrodni"
            elif gender_b == Person.FEMALE:
                rel_str = "siostra przyrodnia"
            else:
                rel_str = "brat lub siostra przyrodnia"
        else:
            rel_str = "nieokreślona relacja rodzeństwa"
        return rel_str


if __name__ == "__main__":
    # Test function. Call it as follows from the command line (so as to find
    #        imported modules):
    #    export PYTHONPATH=/path/to/gramps/src
    #    python src/plugins/rel/rel_pl.py

    """TRANSLATORS, copy this if statement at the bottom of your
    rel_xx.py module, and test your work with:
    python src/plugins/rel/rel_xx.py
    """
    from gramps.gen.relationship import test

    RC = RelationshipCalculator()
    test(RC, True)



# -*- coding: utf-8 -*-
#
# Gramps - a GTK+/GNOME based genealogy program
#
# Copyright (C) 2003-2005  Donald N. Allingham
# Copyright (C) 2008       Stefan Siegel
# Copyright (C) 2008       Brian G. Matherly
# Copyright (C) 2021       Mirko Leonhaeuser
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#

# Original version written by Alex Roitman, largely based on relationship.py
# by Don Allingham and on valuable input from Dr. Martin Senftleben
# Modified by Joachim Breitner to not use „Großcousine“, in accordance with
# http://de.wikipedia.org/wiki/Verwandtschaftsbeziehung
# Rewritten from scratch for Gramps 3 by Stefan Siegel,
# loosely based on rel_fr.py
"""
German-specific classes for relationships.
"""

# -------------------------------------------------------------------------
#
# standard python modules
#
# -------------------------------------------------------------------------

import re

# -------------------------------------------------------------------------
#
# Gramps modules
#
# -------------------------------------------------------------------------

from gramps.gen.lib import Person
import gramps.gen.relationship

# -------------------------------------------------------------------------
#
#
#
# -------------------------------------------------------------------------

_ordinal = [
    "nullte",
    "erste",
    "zweite",
    "dritte",
    "vierte",
    "fünfte",
    "sechste",
    "siebte",
    "achte",
    "neunte",
    "zehnte",
    "elfte",
    "zwölfte",
]

_removed = [
    "",
    "",
    "Groß",
    "Urgroß",
    "Alt",
    "Altgroß",
    "Alturgroß",
    "Ober",
    "Obergroß",
    "Oberurgroß",
    "Stamm",
    "Stammgroß",
    "Stammurgroß",
    "Ahnen",
    "Ahnengroß",
    "Ahnenurgroß",
    "Urahnen",
    "Urahnengroß",
    "Urahnenurgroß",
    "Erz",
    "Erzgroß",
    "Erzurgroß",
    "Erzahnen",
    "Erzahnengroß",
    "Erzahnenurgroß",
]

_lineal_up = {
    "many": "%(p)sEltern%(s)s",
    "unknown": "%(p)sElter%(s)s",  # "Elter" sounds strange but is correct
    "male": "%(p)sVater%(s)s",
    "female": "%(p)sMutter%(s)s",
}
_lineal_down = {
    "many": "%(p)sKinder%(s)s",
    "unknown": "%(p)sKind%(s)s",
    "male": "%(p)sSohn%(s)s",
    "female": "%(p)sTochter%(s)s",
}
_collateral_up = {
    "many": "%(p)sOnkel und %(p)sTanten%(s)s",
    "unknown": "%(p)sOnkel oder %(p)sTante%(s)s",
    "male": "%(p)sOnkel%(s)s",
    "female": "%(p)sTante%(s)s",
}
_collateral_down = {
    "many": "%(p)sNeffen und %(p)sNichten%(s)s",
    "unknown": "%(p)sNeffe oder %(p)sNichte%(s)s",
    "male": "%(p)sNeffe%(s)s",
    "female": "%(p)sNichte%(s)s",
}
_collateral_same = {
    "many": "%(p)sCousins und %(p)sCousinen%(s)s",
    "unknown": "%(p)sCousin oder %(p)sCousine%(s)s",
    "male": "%(p)sCousin%(s)s",
    "female": "%(p)sCousine%(s)s",
}
_collateral_sib = {
    "many": "%(p)sGeschwister%(s)s",
    "unknown": "%(p)sGeschwisterkind%(s)s",
    "male": "%(p)sBruder%(s)s",
    "female": "%(p)sSchwester%(s)s",
}

_schwager = {
    "many": "%(p)sSchwager%(s)s",
    "unknown": "%(p)sSchwager%(s)s",
    "male": "%(p)sSchwager%(s)s",
    "female": "%(p)sSchwägerin%(s)s",
}
_schwippschwager = {
    "many": "%(p)sSchwippschwager%(s)s",
    "unknown": "%(p)sSchwippschwager%(s)s",
    "male": "%(p)sSchwippschwager%(s)s",
    "female": "%(p)sSchwippschwägerin%(s)s",
}

# -------------------------------------------------------------------------
#
#
#
# -------------------------------------------------------------------------


class RelationshipCalculator(gramps.gen.relationship.RelationshipCalculator):
    """
    RelationshipCalculator Class
    """

    def __init__(self):
        gramps.gen.relationship.RelationshipCalculator.__init__(self)

    def _make_roman(self, num):
        roman = ""
        for v, r in [
            (1000, "M"),
            (900, "CM"),
            (500, "D"),
            (400, "CD"),
            (100, "C"),
            (90, "XC"),
            (50, "L"),
            (40, "XL"),
            (10, "X"),
            (9, "IX"),
            (5, "V"),
            (4, "IV"),
            (1, "I"),
        ]:
            while num > v:
                num -= v
                roman += r
        return roman

    def _fix_caps(self, string):
        return re.sub(r"(?<=[^\s(/A-Z])[A-Z]", lambda m: m.group().lower(), string)

    def _removed_text(self, degree, removed):
        if (degree, removed) == (0, -2):
            return "Enkel"
        elif (degree, removed) == (0, -3):
            return "Urenkel"
        removed = abs(removed)
        if removed < len(_removed):
            return _removed[removed]
        else:
            return "(%s) Urgroß" % self._make_roman(removed - 1)

    def _degree_text(self, degree, removed):
        if removed == 0:
            degree -= 1  # a cousin has same degree as his parent (uncle/aunt)
        if degree <= 1:
            return ""
        if degree < len(_ordinal):
            return " %sn Grades" % _ordinal[degree]
        else:
            return " %d. Grades" % degree

    def _gender_convert(self, gender):
        if gender == Person.MALE:
            return "male"
        elif gender == Person.FEMALE:
            return "female"
        else:
            return "unknown"

    def _get_relationship_string(
        self,
        Ga,
        Gb,
        gender,
        reltocommon_a="",
        reltocommon_b="",
        only_birth=True,
        in_law_a=False,
        in_law_b=False,
    ):
        common_ancestor_count = 0
        if reltocommon_a == "":
            reltocommon_a = self.REL_FAM_BIRTH
        if reltocommon_b == "":
            reltocommon_b = self.REL_FAM_BIRTH
        if reltocommon_a[-1] in [
            self.REL_MOTHER,
            self.REL_FAM_BIRTH,
            self.REL_FAM_BIRTH_MOTH_ONLY,
        ] and reltocommon_b[-1] in [
            self.REL_MOTHER,
            self.REL_FAM_BIRTH,
            self.REL_FAM_BIRTH_MOTH_ONLY,
        ]:
            common_ancestor_count += 1  # same female ancestor
        if reltocommon_a[-1] in [
            self.REL_FATHER,
            self.REL_FAM_BIRTH,
            self.REL_FAM_BIRTH_FATH_ONLY,
        ] and reltocommon_b[-1] in [
            self.REL_FATHER,
            self.REL_FAM_BIRTH,
            self.REL_FAM_BIRTH_FATH_ONLY,
        ]:
            common_ancestor_count += 1  # same male ancestor

        degree = min(Ga, Gb)
        removed = Ga - Gb

        if degree == 0 and removed < 0:
            # for descendants the "in-law" logic is reversed
            (in_law_a, in_law_b) = (in_law_b, in_law_a)

        rel_str = ""
        pre = ""
        post = ""

        if in_law_b and degree == 0:
            pre += "Stief"
        elif (not only_birth) or common_ancestor_count == 0:
            pre += "Stief-/Adoptiv"
        if in_law_a and (degree, removed) != (1, 0):
            # A "Schwiegerbruder" really is a "Schwager" (handled later)
            pre += "Schwieger"
        if degree != 0 and common_ancestor_count == 1:
            pre += "Halb"
        pre += self._removed_text(degree, removed)
        post += self._degree_text(degree, removed)
        if in_law_b and degree != 0 and (degree, removed) != (1, 0):
            # A "Bruder (angeheiratet)" also is a "Schwager" (handled later)
            post += " (angeheiratet)"

        if degree == 0:
            # lineal relationship
            if removed > 0:
                rel_str = _lineal_up[gender]
            elif removed < 0:
                rel_str = _lineal_down[gender]
            elif in_law_a or in_law_b:
                rel_str = "Partner"
            else:
                rel_str = "Proband"
        else:
            # collateral relationship
            if removed > 0:
                rel_str = _collateral_up[gender]
            elif removed < 0:
                rel_str = _collateral_down[gender]
            elif degree == 1:
                if in_law_a or in_law_b:
                    if in_law_a and in_law_b:
                        rel_str = _schwippschwager[gender]
                    else:
                        rel_str = _schwager[gender]
                else:
                    rel_str = _collateral_sib[gender]
            else:
                rel_str = _collateral_same[gender]
        return self._fix_caps(rel_str % {"p": pre, "s": post})

    def get_plural_relationship_string(
        self,
        Ga,
        Gb,
        reltocommon_a="",
        reltocommon_b="",
        only_birth=True,
        in_law_a=False,
        in_law_b=False,
    ):
        return self._get_relationship_string(
            Ga, Gb, "many", reltocommon_a, reltocommon_b, only_birth, in_law_a, in_law_b
        )

    def get_single_relationship_string(
        self,
        Ga,
        Gb,
        gender_a,
        gender_b,
        reltocommon_a,
        reltocommon_b,
        only_birth=True,
        in_law_a=False,
        in_law_b=False,
    ):
        return self._get_relationship_string(
            Ga,
            Gb,
            self._gender_convert(gender_b),
            reltocommon_a,
            reltocommon_b,
            only_birth,
            in_law_a,
            in_law_b,
        )

    def get_sibling_relationship_string(
        self, sib_type, gender_a, gender_b, in_law_a=False, in_law_b=False
    ):
        if sib_type in [self.NORM_SIB, self.UNKNOWN_SIB]:
            # the NORM_SIB translation is generic and suitable for UNKNOWN_SIB
            rel = self.REL_FAM_BIRTH
            only_birth = True
        elif sib_type == self.HALF_SIB_FATHER:
            rel = self.REL_FAM_BIRTH_FATH_ONLY
            only_birth = True
        elif sib_type == self.HALF_SIB_MOTHER:
            rel = self.REL_FAM_BIRTH_MOTH_ONLY
            only_birth = True
        elif sib_type == self.STEP_SIB:
            rel = self.REL_FAM_NONBIRTH
            only_birth = False
        return self._get_relationship_string(
            1,
            1,
            self._gender_convert(gender_b),
            rel,
            rel,
            only_birth,
            in_law_a,
            in_law_b,
        )


if __name__ == "__main__":
    # Test function. Call it as follows from the command line (so as to find
    #        imported modules):
    #    export PYTHONPATH=/path/to/gramps/src
    # python src/plugins/rel/rel_de.py
    # (Above not needed here)

    """TRANSLATORS, copy this if statement at the bottom of your
    rel_xx.py module, and test your work with:
    python src/plugins/rel/rel_xx.py
    """
    from gramps.gen.relationship import test

    rc = RelationshipCalculator()
    test(rc, True)
